import itertools
from types import (
    TracebackType,
)
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

from cached_property import (
    cached_property,
)
from eth_typing import (
    Address,
)
from eth_utils import (
    encode_hex,
    get_extended_debug_logger,
)

from eth._utils.datatypes import (
    Configurable,
)
from eth._utils.numeric import (
    ceil32,
)
from eth.abc import (
    CodeStreamAPI,
    ComputationAPI,
    GasMeterAPI,
    MemoryAPI,
    MessageAPI,
    OpcodeAPI,
    StackAPI,
    StateAPI,
    TransactionContextAPI,
)
from eth.constants import (
    GAS_MEMORY,
    GAS_MEMORY_QUADRATIC_DENOMINATOR,
)
from eth.exceptions import (
    Halt,
    VMError,
)
from eth.typing import (
    BytesOrView,
)
from eth.validation import (
    validate_canonical_address,
    validate_is_bytes,
    validate_uint256,
)
from eth.vm.code_stream import (
    CodeStream,
)
from eth.vm.gas_meter import (
    GasMeter,
)
from eth.vm.logic.invalid import (
    InvalidOpcode,
)
from eth.vm.memory import (
    Memory,
)
from eth.vm.message import (
    Message,
)
from eth.vm.stack import (
    Stack,
)


def NO_RESULT(computation: ComputationAPI) -> None:
    """
    This is a special method intended for usage as the "no precompile found" result.
    The type signature is designed to match the other precompiles.
    """
    raise Exception("This method is never intended to be executed")


def memory_gas_cost(size_in_bytes: int) -> int:
    size_in_words = ceil32(size_in_bytes) // 32
    linear_cost = size_in_words * GAS_MEMORY
    quadratic_cost = size_in_words**2 // GAS_MEMORY_QUADRATIC_DENOMINATOR

    total_cost = linear_cost + quadratic_cost
    return total_cost


class BaseComputation(ComputationAPI, Configurable):
    """
    The base class for all execution computations.

      .. note::

        Each :class:`~eth.vm.computation.BaseComputation` class must be configured with:

        ``opcodes``:        A mapping from the opcode integer value to the logic
                            function for the opcode.

        ``_precompiles``:   A mapping of contract address to the precompile function
                            for execution of precompiled contracts.
    """

    logger = get_extended_debug_logger("eth.vm.computation.BaseComputation")

    state: StateAPI = None
    msg: MessageAPI = None
    transaction_context: TransactionContextAPI = None
    code: CodeStreamAPI = None
    children: List[ComputationAPI] = None
    return_data: bytes = b""
    accounts_to_delete: Dict[Address, Address] = None

    invalid_opcode_class: Type[OpcodeAPI] = InvalidOpcode

    _memory: MemoryAPI = None
    _stack: StackAPI = None
    _gas_meter: GasMeterAPI = None
    _error: VMError = None
    _output: bytes = b""
    _log_entries: List[Tuple[int, Address, Tuple[int, ...], bytes]] = None

    # VM configuration
    opcodes: Dict[int, OpcodeAPI] = None
    _precompiles: Dict[Address, Callable[[ComputationAPI], ComputationAPI]] = None

    def __init__(
        self,
        state: StateAPI,
        message: MessageAPI,
        transaction_context: TransactionContextAPI,
    ) -> None:
        self.state = state
        self.msg = message
        self.transaction_context = transaction_context
        self.code = CodeStream(message.code)

        self._gas_meter = self._configure_gas_meter()

        self.children = []
        self.accounts_to_delete = {}
        self._stack = Stack()
        self._memory = Memory()
        self._log_entries = []

    def _configure_gas_meter(self) -> GasMeter:
        return GasMeter(self.msg.gas)

    # -- class methods -- #
    @classmethod
    def apply_message(
        cls,
        state: StateAPI,
        message: MessageAPI,
        transaction_context: TransactionContextAPI,
    ) -> ComputationAPI:
        raise NotImplementedError("Must be implemented by subclasses")

    @classmethod
    def apply_create_message(
        cls,
        state: StateAPI,
        message: MessageAPI,
        transaction_context: TransactionContextAPI,
    ) -> ComputationAPI:
        raise NotImplementedError("Must be implemented by subclasses")

    # -- convenience -- #
    @property
    def is_origin_computation(self) -> bool:
        return self.msg.sender == self.transaction_context.origin

    # -- runtime operations -- #
    def prepare_child_message(
        self,
        gas: int,
        to: Address,
        value: int,
        data: BytesOrView,
        code: bytes,
        **kwargs: Any,
    ) -> MessageAPI:
        kwargs.setdefault("sender", self.msg.storage_address)

        child_message = Message(
            gas=gas,
            to=to,
            value=value,
            data=data,
            code=code,
            depth=self.msg.depth + 1,
            **kwargs,
        )
        return child_message

    def apply_child_computation(
        self,
        child_msg: MessageAPI,
    ) -> ComputationAPI:
        child_computation = self.generate_child_computation(child_msg)
        self.add_child_computation(child_computation)
        return child_computation

    def generate_child_computation(
        self,
        child_msg: MessageAPI,
    ) -> ComputationAPI:
        if child_msg.is_create:
            child_computation = self.apply_create_message(
                self.state,
                child_msg,
                self.transaction_context,
            )
        else:
            child_computation = self.apply_message(
                self.state,
                child_msg,
                self.transaction_context,
            )
        return child_computation

    def add_child_computation(
        self,
        child_computation: ComputationAPI,
    ) -> None:
        if child_computation.is_error:
            if child_computation.msg.is_create:
                self.return_data = child_computation.output
            elif child_computation.should_burn_gas:
                self.return_data = b""
            else:
                self.return_data = child_computation.output
        else:
            if child_computation.msg.is_create:
                self.return_data = b""
            else:
                self.return_data = child_computation.output
        self.children.append(child_computation)

    # -- gas consumption -- #
    def get_gas_refund(self) -> int:
        if self.is_error:
            return 0
        else:
            return self._gas_meter.gas_refunded + sum(
                c.get_gas_refund() for c in self.children
            )

    # -- account management -- #
    def register_account_for_deletion(self, beneficiary: Address) -> None:
        # SELFDESTRUCT

        validate_canonical_address(
            beneficiary,
            title="Self destruct beneficiary address",
        )

        if self.msg.storage_address in self.accounts_to_delete:
            raise ValueError(
                "Invariant.  Should be impossible for an account to be "
                "registered for deletion multiple times"
            )
        self.accounts_to_delete[self.msg.storage_address] = beneficiary

    def get_accounts_for_deletion(self) -> Tuple[Tuple[Address, Address], ...]:
        # SELFDESTRUCT

        if self.is_error:
            return ()
        else:
            return tuple(
                dict(
                    itertools.chain(
                        self.accounts_to_delete.items(),
                        *(child.get_accounts_for_deletion() for child in self.children),
                    )
                ).items()
            )

    # -- EVM logging -- #
    def add_log_entry(
        self,
        account: Address,
        topics: Tuple[int, ...],
        data: bytes,
    ) -> None:
        validate_canonical_address(account, title="Log entry address")
        for topic in topics:
            validate_uint256(topic, title="Log entry topic")
        validate_is_bytes(data, title="Log entry data")
        self._log_entries.append(
            (self.transaction_context.get_next_log_counter(), account, topics, data)
        )

    def get_raw_log_entries(
        self,
    ) -> Tuple[Tuple[int, bytes, Tuple[int, ...], bytes], ...]:
        if self.is_error:
            return ()
        else:
            return tuple(
                sorted(
                    itertools.chain(
                        self._log_entries,
                        *(child.get_raw_log_entries() for child in self.children),
                    )
                )
            )

    def get_log_entries(self) -> Tuple[Tuple[bytes, Tuple[int, ...], bytes], ...]:
        return tuple(log[1:] for log in self.get_raw_log_entries())

    # -- state transition -- #
    @classmethod
    def apply_computation(
        cls,
        state: StateAPI,
        message: MessageAPI,
        transaction_context: TransactionContextAPI,
    ) -> ComputationAPI:
        with cls(state, message, transaction_context) as computation:
            if message.is_create and computation.is_origin_computation:
                # If computation is from a create transaction, consume initcode gas if
                # >= Shanghai. CREATE and CREATE2 are handled in the opcode
                # implementations.
                cls.consume_initcode_gas_cost(computation)

            # Early exit on pre-compiles
            precompile = computation.precompiles.get(message.code_address, NO_RESULT)
            if precompile is not NO_RESULT:
                precompile(computation)
                return computation

            show_debug2 = computation.logger.show_debug2

            opcode_lookup = computation.opcodes
            for opcode in computation.code:
                try:
                    opcode_fn = opcode_lookup[opcode]
                except KeyError:
                    opcode_fn = computation.invalid_opcode_class(opcode)

                if show_debug2:
                    # We dig into some internals for debug logs
                    base_comp = cast(BaseComputation, computation)
                    computation.logger.debug2(
                        "OPCODE: 0x%x (%s) | pc: %s | stack: %s",
                        opcode,
                        opcode_fn.mnemonic,
                        max(0, computation.code.program_counter - 1),
                        base_comp._stack,
                    )

                try:
                    opcode_fn(computation=computation)
                except Halt:
                    break

        return computation

    # -- error handling -- #
    @property
    def is_success(self) -> bool:
        return self._error is None

    @property
    def is_error(self) -> bool:
        return not self.is_success

    @property
    def error(self) -> VMError:
        if self._error is not None:
            return self._error
        raise AttributeError("Computation does not have an error")

    @error.setter
    def error(self, value: VMError) -> None:
        if self._error is not None:
            raise AttributeError(f"Computation already has an error set: {self._error}")
        self._error = value

    def raise_if_error(self) -> None:
        if self._error is not None:
            raise self._error

    @property
    def should_burn_gas(self) -> bool:
        return self.is_error and self._error.burns_gas

    @property
    def should_return_gas(self) -> bool:
        return not self.should_burn_gas

    @property
    def should_erase_return_data(self) -> bool:
        return self.is_error and self._error.erases_return_data

    # -- memory management -- #
    def extend_memory(self, start_position: int, size: int) -> None:
        validate_uint256(start_position, title="Memory start position")
        validate_uint256(size, title="Memory size")

        before_size = ceil32(len(self._memory))
        after_size = ceil32(start_position + size)

        before_cost = memory_gas_cost(before_size)
        after_cost = memory_gas_cost(after_size)

        if self.logger.show_debug2:
            self.logger.debug2(
                "MEMORY: size (%s -> %s) | cost (%s -> %s)",
                before_size,
                after_size,
                before_cost,
                after_cost,
            )

        if size:
            if before_cost < after_cost:
                gas_fee = after_cost - before_cost
                self._gas_meter.consume_gas(
                    gas_fee,
                    reason=" ".join(
                        (
                            "Expanding memory",
                            str(before_size),
                            "->",
                            str(after_size),
                        )
                    ),
                )

            self._memory.extend(start_position, size)

    def memory_write(self, start_position: int, size: int, value: bytes) -> None:
        return self._memory.write(start_position, size, value)

    def memory_read(self, start_position: int, size: int) -> memoryview:
        return self._memory.read(start_position, size)

    def memory_read_bytes(self, start_position: int, size: int) -> bytes:
        return self._memory.read_bytes(start_position, size)

    # -- gas consumption -- #
    def get_gas_meter(self) -> GasMeterAPI:
        return self._gas_meter

    def consume_gas(self, amount: int, reason: str) -> None:
        return self._gas_meter.consume_gas(amount, reason)

    def return_gas(self, amount: int) -> None:
        return self._gas_meter.return_gas(amount)

    def refund_gas(self, amount: int) -> None:
        return self._gas_meter.refund_gas(amount)

    def get_gas_used(self) -> int:
        if self.should_burn_gas:
            return self._gas_meter.start_gas
        else:
            return max(
                0,
                self._gas_meter.start_gas - self._gas_meter.gas_remaining,
            )

    def get_gas_remaining(self) -> int:
        if self.should_burn_gas:
            return 0
        else:
            return self._gas_meter.gas_remaining

    @classmethod
    def consume_initcode_gas_cost(cls, computation: ComputationAPI) -> None:
        # this method does not become relevant until the Shanghai hard fork
        """
        Before starting the computation, consume initcode gas cost.
        """
        pass

    # -- stack management -- #
    def stack_swap(self, position: int) -> None:
        return self._stack.swap(position)

    def stack_dup(self, position: int) -> None:
        return self._stack.dup(position)

    # Stack manipulation is performance-sensitive code.
    # Avoid method call overhead by proxying stack method directly to stack object

    @cached_property
    def stack_pop_ints(self) -> Callable[[int], Tuple[int, ...]]:
        return self._stack.pop_ints

    @cached_property
    def stack_pop_bytes(self) -> Callable[[int], Tuple[bytes, ...]]:
        return self._stack.pop_bytes

    @cached_property
    def stack_pop_any(self) -> Callable[[int], Tuple[Union[int, bytes], ...]]:
        return self._stack.pop_any

    @cached_property
    def stack_pop1_int(self) -> Callable[[], int]:
        return self._stack.pop1_int

    @cached_property
    def stack_pop1_bytes(self) -> Callable[[], bytes]:
        return self._stack.pop1_bytes

    @cached_property
    def stack_pop1_any(self) -> Callable[[], Union[int, bytes]]:
        return self._stack.pop1_any

    @cached_property
    def stack_push_int(self) -> Callable[[int], None]:
        return self._stack.push_int

    @cached_property
    def stack_push_bytes(self) -> Callable[[bytes], None]:
        return self._stack.push_bytes

    # -- computation result -- #
    @property
    def output(self) -> bytes:
        if self.should_erase_return_data:
            return b""
        else:
            return self._output

    @output.setter
    def output(self, value: bytes) -> None:
        validate_is_bytes(value)
        self._output = value

    # -- opcode API -- #
    @property
    def precompiles(self) -> Dict[Address, Callable[[ComputationAPI], Any]]:
        if self._precompiles is None:
            return {}
        else:
            return self._precompiles

    @classmethod
    def get_precompiles(cls) -> Dict[Address, Callable[[ComputationAPI], Any]]:
        if cls._precompiles is None:
            return {}
        else:
            return cls._precompiles

    def get_opcode_fn(self, opcode: int) -> OpcodeAPI:
        try:
            return self.opcodes[opcode]
        except KeyError:
            return InvalidOpcode(opcode)

    # -- context manager API -- #
    def __enter__(self) -> ComputationAPI:
        if self.logger.show_debug2:
            self.logger.debug2(
                (
                    "MESSAGE COMPUTATION STARTING: "
                    "from: %s | to: %s | value: %s | depth %s | static: %s | gas: %s"
                ),
                encode_hex(self.msg.sender),
                encode_hex(self.msg.to),
                self.msg.value,
                self.msg.depth,
                "y" if self.msg.is_static else "n",
                self.msg.gas,
            )
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Union[None, bool]:
        if exc_value and isinstance(exc_value, VMError):
            if self.logger.show_debug2:
                self.logger.debug2(
                    (
                        "COMPUTATION ERROR: "
                        "gas: %s | from: %s | to: %s | value: %s | "
                        "depth: %s | static: %s | error: %s"
                    ),
                    self.msg.gas,
                    encode_hex(self.msg.sender),
                    encode_hex(self.msg.to),
                    self.msg.value,
                    self.msg.depth,
                    "y" if self.msg.is_static else "n",
                    exc_value,
                )

            self._error = exc_value
            if self.should_burn_gas:
                self.consume_gas(
                    self._gas_meter.gas_remaining,
                    reason=" ".join(
                        (
                            "Zeroing gas due to VM Exception:",
                            str(exc_value),
                        )
                    ),
                )

            # when we raise an exception that erases return data, erase the return data
            if self.should_erase_return_data:
                self.return_data = b""

            # suppress VM exceptions
            return True

        elif exc_type is None and self.logger.show_debug2:
            self.logger.debug2(
                (
                    "COMPUTATION SUCCESS: "
                    "from: %s | to: %s | value: %s | depth: %s | static: %s "
                    "| gas-used: %s | gas-remaining: %s"
                ),
                encode_hex(self.msg.sender),
                encode_hex(self.msg.to),
                self.msg.value,
                self.msg.depth,
                "y" if self.msg.is_static else "n",
                self.get_gas_used(),
                self._gas_meter.gas_remaining,
            )

        return None



from functools import (
    partial,
)
from typing import (
    Tuple,
)

from eth_keys.datatypes import (
    PrivateKey,
)
from eth_typing import (
    Address,
)
import rlp

from eth._utils.transactions import (
    V_OFFSET,
    IntrinsicGasSchedule,
    calculate_intrinsic_gas,
    create_transaction_signature,
    extract_transaction_sender,
    validate_transaction_signature,
)
from eth.abc import (
    ReceiptAPI,
    SignedTransactionAPI,
)
from eth.constants import (
    CREATE_CONTRACT_ADDRESS,
    GAS_TX,
    GAS_TXDATANONZERO,
    GAS_TXDATAZERO,
)
from eth.rlp.logs import (
    Log,
)
from eth.rlp.receipts import (
    Receipt,
)
from eth.rlp.transactions import (
    BaseTransaction,
    BaseUnsignedTransaction,
)
from eth.validation import (
    validate_canonical_address,
    validate_gte,
    validate_is_bytes,
    validate_is_integer,
    validate_lt_secpk1n,
    validate_lte,
    validate_uint64,
    validate_uint256,
)

FRONTIER_TX_GAS_SCHEDULE = IntrinsicGasSchedule(
    gas_tx=GAS_TX,
    gas_txcreate=0,
    gas_txdatazero=GAS_TXDATAZERO,
    gas_txdatanonzero=GAS_TXDATANONZERO,
)


frontier_get_intrinsic_gas = partial(calculate_intrinsic_gas, FRONTIER_TX_GAS_SCHEDULE)


class FrontierTransaction(BaseTransaction):
    @property
    def y_parity(self) -> int:
        return self.v - V_OFFSET

    @property
    def v_min(self) -> int:
        return V_OFFSET

    @property
    def v_max(self) -> int:
        return V_OFFSET + 1

    def validate(self) -> None:
        validate_uint64(self.nonce, title="Transaction.nonce")
        validate_uint256(self.gas_price, title="Transaction.gas_price")
        validate_uint256(self.gas, title="Transaction.gas")
        if self.to != CREATE_CONTRACT_ADDRESS:
            validate_canonical_address(self.to, title="Transaction.to")
        validate_uint256(self.value, title="Transaction.value")
        validate_is_bytes(self.data, title="Transaction.data")

        validate_uint256(self.v, title="Transaction.v")
        validate_uint256(self.r, title="Transaction.r")
        validate_uint256(self.s, title="Transaction.s")

        validate_lt_secpk1n(self.r, title="Transaction.r")
        validate_gte(self.r, minimum=1, title="Transaction.r")
        validate_lt_secpk1n(self.s, title="Transaction.s")
        validate_gte(self.s, minimum=1, title="Transaction.s")

        validate_gte(self.v, minimum=self.v_min, title="Transaction.v")
        validate_lte(self.v, maximum=self.v_max, title="Transaction.v")

        super().validate()

    def check_signature_validity(self) -> None:
        validate_transaction_signature(self)

    def get_sender(self) -> Address:
        return extract_transaction_sender(self)

    def get_intrinsic_gas(self) -> int:
        return frontier_get_intrinsic_gas(self)

    def get_message_for_signing(self) -> bytes:
        return rlp.encode(
            FrontierUnsignedTransaction(
                nonce=self.nonce,
                gas_price=self.gas_price,
                gas=self.gas,
                to=self.to,
                value=self.value,
                data=self.data,
            )
        )

    @classmethod
    def create_unsigned_transaction(
        cls,
        *,
        nonce: int,
        gas_price: int,
        gas: int,
        to: Address,
        value: int,
        data: bytes
    ) -> "FrontierUnsignedTransaction":
        return FrontierUnsignedTransaction(nonce, gas_price, gas, to, value, data)

    @classmethod
    def new_transaction(
        cls,
        nonce: int,
        gas_price: int,
        gas: int,
        to: Address,
        value: int,
        data: bytes,
        v: int,
        r: int,
        s: int,
    ) -> SignedTransactionAPI:
        return cls(nonce, gas_price, gas, to, value, data, v, r, s)

    def make_receipt(
        self,
        status: bytes,
        gas_used: int,
        log_entries: Tuple[Tuple[bytes, Tuple[int, ...], bytes], ...],
    ) -> ReceiptAPI:
        # 'status' is a misnomer in Frontier. Until Byzantium, it is the
        # intermediate state root.

        logs = [Log(address, topics, data) for address, topics, data in log_entries]

        return Receipt(
            state_root=status,
            gas_used=gas_used,
            logs=logs,
        )

    # Old transactions are treated as setting both max-fees as the gas price
    @property
    def max_priority_fee_per_gas(self) -> int:
        return self.gas_price

    @property
    def max_fee_per_gas(self) -> int:
        return self.gas_price


class FrontierUnsignedTransaction(BaseUnsignedTransaction):
    def validate(self) -> None:
        validate_uint64(self.nonce, title="Transaction.nonce")
        validate_is_integer(self.gas_price, title="Transaction.gas_price")
        validate_uint256(self.gas, title="Transaction.gas")
        if self.to != CREATE_CONTRACT_ADDRESS:
            validate_canonical_address(self.to, title="Transaction.to")
        validate_uint256(self.value, title="Transaction.value")
        validate_is_bytes(self.data, title="Transaction.data")
        super().validate()

    def as_signed_transaction(
        self,
        private_key: PrivateKey,
        chain_id: int = None,  # unused until SpuriousDragon
    ) -> FrontierTransaction:
        v, r, s = create_transaction_signature(self, private_key)
        return FrontierTransaction(
            nonce=self.nonce,
            gas_price=self.gas_price,
            gas=self.gas,
            to=self.to,
            value=self.value,
            data=self.data,
            v=v,
            r=r,
            s=s,
        )

    def get_intrinsic_gas(self) -> int:
        return frontier_get_intrinsic_gas(self)

    # Old transactions are treated as setting both max-fees as the gas price
    @property
    def max_priority_fee_per_gas(self) -> int:
        return self.gas_price

    @property
    def max_fee_per_gas(self) -> int:
        return self.gas_price



from typing import (
    cast,
)

from cached_property import (
    cached_property,
)
from eth_typing import (
    Address,
    Hash32,
)
from eth_utils import (
    keccak,
)
import rlp
from rlp.sedes import (
    big_endian_int,
)

from eth.abc import (
    WithdrawalAPI,
)
from eth.rlp.sedes import (
    address,
)
from eth.validation import (
    validate_canonical_address,
    validate_uint64,
)


class Withdrawal(rlp.Serializable):
    fields = [
        ("index", big_endian_int),
        ("validator_index", big_endian_int),
        ("address", address),
        ("amount", big_endian_int),
    ]

    def __init__(
        self,
        index: int = 0,
        validator_index: int = 0,
        address: Address = None,
        amount: int = 0,
    ) -> None:
        super().__init__(
            index=index,
            validator_index=validator_index,
            address=address,
            amount=amount,
        )

    def validate(self) -> None:
        validate_uint64(self.index, "Withdrawal.index")
        validate_uint64(self.validator_index, "Withdrawal.validator_index")
        validate_canonical_address(self.address, "Withdrawal.address")
        validate_uint64(self.amount, "Withdrawal.amount")

    @classmethod
    def decode(cls, encoded: bytes) -> WithdrawalAPI:
        return rlp.decode(encoded, sedes=cls)

    def encode(self) -> bytes:
        return rlp.encode(self)

    @cached_property
    def hash(self) -> Hash32:
        return cast(Hash32, keccak(self.encode()))



from eth.abc import (
    ComputationAPI,
)


def mstore(computation: ComputationAPI) -> None:
    start_position = computation.stack_pop1_int()
    value = computation.stack_pop1_bytes()

    padded_value = value.rjust(32, b"\x00")
    normalized_value = padded_value[-32:]

    computation.extend_memory(start_position, 32)

    computation.memory_write(start_position, 32, normalized_value)


def mstore8(computation: ComputationAPI) -> None:
    start_position = computation.stack_pop1_int()
    value = computation.stack_pop1_bytes()

    padded_value = value.rjust(1, b"\x00")
    normalized_value = padded_value[-1:]

    computation.extend_memory(start_position, 1)

    computation.memory_write(start_position, 1, normalized_value)


def mload(computation: ComputationAPI) -> None:
    start_position = computation.stack_pop1_int()

    computation.extend_memory(start_position, 32)

    value = computation.memory_read_bytes(start_position, 32)
    computation.stack_push_bytes(value)


def msize(computation: ComputationAPI) -> None:
    computation.stack_push_int(len(computation._memory))



from abc import (
    ABC,
)
from typing import (
    Type,
)

from eth_utils import (
    encode_hex,
)
from rlp.sedes import (
    CountableList,
)

from eth.abc import (
    TransactionBuilderAPI,
)

from ..arrow_glacier import (
    ArrowGlacierBlock,
)
from ..arrow_glacier.blocks import (
    ArrowGlacierBlockHeader,
    ArrowGlacierMiningHeader,
)
from ..london.blocks import (
    LondonBackwardsHeader,
)
from .transactions import (
    GrayGlacierTransactionBuilder,
)


class GrayGlacierMiningHeader(ArrowGlacierMiningHeader, ABC):
    pass


class GrayGlacierBlockHeader(ArrowGlacierBlockHeader, ABC):
    def __str__(self) -> str:
        return f"<GrayGlacierBlockHeader #{self.block_number} {encode_hex(self.hash)[2:10]}>"  # noqa: E501


class GrayGlacierBlock(ArrowGlacierBlock):
    transaction_builder: Type[TransactionBuilderAPI] = GrayGlacierTransactionBuilder
    fields = [
        ("header", GrayGlacierBlockHeader),
        ("transactions", CountableList(transaction_builder)),
        ("uncles", CountableList(LondonBackwardsHeader)),
    ]



from eth.vm.forks.shanghai.blocks import (
    ShanghaiBackwardsHeader,
)

HeaderSedes = ShanghaiBackwardsHeader
"""
An RLP codec that can decode *all* known header types.

Unfortunately, we often cannot look up the VM to determine the valid codec. For
example, when looking up a header by hash, we don't have the block number yet,
and so can't load the VM configuration to find out which VM's rules to use to
decode the header. Also, it's useful to have this universal sedes class when
decoding multiple uncles that span a fork block.
"""



import hashlib

from eth import (
    constants,
)
from eth._utils.numeric import (
    ceil32,
)
from eth.abc import (
    ComputationAPI,
)


def sha256(computation: ComputationAPI) -> ComputationAPI:
    word_count = ceil32(len(computation.msg.data)) // 32
    gas_fee = constants.GAS_SHA256 + word_count * constants.GAS_SHA256WORD

    computation.consume_gas(gas_fee, reason="SHA256 Precompile")
    input_bytes = computation.msg.data
    hash = hashlib.sha256(input_bytes).digest()
    computation.output = hash
    return computation



import copy
import functools
from typing import (
    Any,
    Callable,
    Dict,
)

from eth_utils.toolz import (
    merge,
)

from eth import (
    constants,
)
from eth.abc import (
    ComputationAPI,
    OpcodeAPI,
)
from eth.exceptions import (
    WriteProtection,
)
from eth.vm import (
    mnemonics,
    opcode_values,
)
from eth.vm.forks.spurious_dragon.opcodes import (
    SPURIOUS_DRAGON_OPCODES,
)
from eth.vm.forks.tangerine_whistle.constants import (
    GAS_CALL_EIP150,
    GAS_SELFDESTRUCT_EIP150,
)
from eth.vm.logic import (
    call,
    context,
    logging,
    storage,
    system,
)
from eth.vm.opcode import (
    as_opcode,
)


def ensure_no_static(opcode_fn: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(opcode_fn)
    def inner(computation: ComputationAPI) -> Callable[..., Any]:
        if computation.msg.is_static:
            raise WriteProtection(
                "Cannot modify state while inside of a STATICCALL context"
            )
        return opcode_fn(computation)

    return inner


UPDATED_OPCODES = {
    opcode_values.REVERT: as_opcode(
        logic_fn=system.revert,
        mnemonic=mnemonics.REVERT,
        gas_cost=constants.GAS_ZERO,
    ),
    #
    # Context
    #
    opcode_values.RETURNDATASIZE: as_opcode(
        logic_fn=context.returndatasize,
        mnemonic=mnemonics.RETURNDATASIZE,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.RETURNDATACOPY: as_opcode(
        logic_fn=context.returndatacopy,
        mnemonic=mnemonics.RETURNDATACOPY,
        gas_cost=constants.GAS_VERYLOW,
    ),
    #
    # Call
    #
    opcode_values.STATICCALL: call.StaticCall.configure(
        __name__="opcode:STATICCALL",
        mnemonic=mnemonics.STATICCALL,
        gas_cost=GAS_CALL_EIP150,
    )(),
    opcode_values.CALL: call.CallByzantium.configure(
        __name__="opcode:CALL",
        mnemonic=mnemonics.CALL,
        gas_cost=GAS_CALL_EIP150,
    )(),
    #
    # Logging
    #
    opcode_values.LOG0: as_opcode(
        logic_fn=ensure_no_static(logging.log0),
        mnemonic=mnemonics.LOG0,
        gas_cost=constants.GAS_LOG,
    ),
    opcode_values.LOG1: as_opcode(
        logic_fn=ensure_no_static(logging.log1),
        mnemonic=mnemonics.LOG1,
        gas_cost=constants.GAS_LOG,
    ),
    opcode_values.LOG2: as_opcode(
        logic_fn=ensure_no_static(logging.log2),
        mnemonic=mnemonics.LOG2,
        gas_cost=constants.GAS_LOG,
    ),
    opcode_values.LOG3: as_opcode(
        logic_fn=ensure_no_static(logging.log3),
        mnemonic=mnemonics.LOG3,
        gas_cost=constants.GAS_LOG,
    ),
    opcode_values.LOG4: as_opcode(
        logic_fn=ensure_no_static(logging.log4),
        mnemonic=mnemonics.LOG4,
        gas_cost=constants.GAS_LOG,
    ),
    #
    # Create
    #
    opcode_values.CREATE: system.CreateByzantium.configure(
        __name__="opcode:CREATE",
        mnemonic=mnemonics.CREATE,
        gas_cost=constants.GAS_CREATE,
    )(),
    #
    # Storage
    #
    opcode_values.SSTORE: as_opcode(
        logic_fn=ensure_no_static(storage.sstore),
        mnemonic=mnemonics.SSTORE,
        gas_cost=constants.GAS_NULL,
    ),
    #
    # Self Destruct
    #
    opcode_values.SELFDESTRUCT: as_opcode(
        logic_fn=ensure_no_static(system.selfdestruct_eip161),
        mnemonic=mnemonics.SELFDESTRUCT,
        gas_cost=GAS_SELFDESTRUCT_EIP150,
    ),
}


BYZANTIUM_OPCODES: Dict[int, OpcodeAPI] = merge(
    copy.deepcopy(SPURIOUS_DRAGON_OPCODES),
    UPDATED_OPCODES,
)



from typing import (
    Optional,
)

from eth.abc import (
    ChainContextAPI,
)
from eth.validation import (
    validate_uint256,
)


class ChainContext(ChainContextAPI):
    __slots__ = ["_chain_id"]

    def __init__(self, chain_id: Optional[int]) -> None:
        if chain_id is None:
            chain_id = 0  # Default value (invalid for public networks)
        # Due to EIP-155's definition of Chain ID,
        # the number that needs to be RLP encoded is `CHAINID * 2 + 36`
        validate_uint256(chain_id)
        self._chain_id = chain_id

    @property
    def chain_id(self) -> int:
        return self._chain_id



from eth_utils import (
    decode_hex,
)

ALLOWED_CLIQUE_DIFFICULTIES = {1, 2}

COMMON_ADDRESS_LENGTH = 20

EPOCH_LENGTH = 30000

IN_MEMORY_SNAPSHOTS = 128

NONCE_AUTH = decode_hex("0xffffffffffffffff")
NONCE_DROP = decode_hex("0x0000000000000000")

# Indicate the byte length required to carry a signature with recovery id.
# 64 bytes ECDSA signature + 1 byte recovery id
SIGNATURE_LENGTH = 64 + 1

# Indicate the byte length reserved as vanity space in extra_data
VANITY_LENGTH = 32



from typing import (
    Optional,
)

from eth_keys.datatypes import (
    PrivateKey,
)
from eth_typing import (
    Address,
)
from eth_utils import (
    int_to_big_endian,
)
import rlp

from eth._utils.numeric import (
    is_even,
)
from eth._utils.transactions import (
    create_transaction_signature,
    extract_chain_id,
    is_eip_155_signed_transaction,
)
from eth.vm.forks.homestead.transactions import (
    HomesteadTransaction,
    HomesteadUnsignedTransaction,
)


class SpuriousDragonTransaction(HomesteadTransaction):
    @property
    def y_parity(self) -> int:
        if is_eip_155_signed_transaction(self):
            return int(is_even(self.v))
        else:
            return super().y_parity

    def get_message_for_signing(self) -> bytes:
        if is_eip_155_signed_transaction(self):
            txn_parts = rlp.decode(rlp.encode(self))
            txn_parts_for_signing = txn_parts[:-3] + [
                int_to_big_endian(self.chain_id),
                b"",
                b"",
            ]
            return rlp.encode(txn_parts_for_signing)
        else:
            return rlp.encode(
                SpuriousDragonUnsignedTransaction(
                    nonce=self.nonce,
                    gas_price=self.gas_price,
                    gas=self.gas,
                    to=self.to,
                    value=self.value,
                    data=self.data,
                )
            )

    @classmethod
    def create_unsigned_transaction(
        cls,
        *,
        nonce: int,
        gas_price: int,
        gas: int,
        to: Address,
        value: int,
        data: bytes
    ) -> "SpuriousDragonUnsignedTransaction":
        return SpuriousDragonUnsignedTransaction(nonce, gas_price, gas, to, value, data)

    @property
    def chain_id(self) -> Optional[int]:
        if is_eip_155_signed_transaction(self):
            return extract_chain_id(self.v)
        else:
            return None

    @property
    def v_min(self) -> int:
        if is_eip_155_signed_transaction(self):
            return 35 + (2 * self.chain_id)
        else:
            return 27

    @property
    def v_max(self) -> int:
        if is_eip_155_signed_transaction(self):
            return 36 + (2 * self.chain_id)
        else:
            return 28


class SpuriousDragonUnsignedTransaction(HomesteadUnsignedTransaction):
    def as_signed_transaction(
        self, private_key: PrivateKey, chain_id: int = None
    ) -> SpuriousDragonTransaction:
        v, r, s = create_transaction_signature(self, private_key, chain_id=chain_id)
        return SpuriousDragonTransaction(
            nonce=self.nonce,
            gas_price=self.gas_price,
            gas=self.gas,
            to=self.to,
            value=self.value,
            data=self.data,
            v=v,
            r=r,
            s=s,
        )



from eth_utils import (
    denoms,
)

GAS_EXTCODEHASH_EIP1052 = 400
GAS_SSTORE_EIP1283_NOOP = 200
GAS_SSTORE_EIP1283_INIT = 20000
GAS_SSTORE_EIP1283_CLEAN = 5000
GAS_SSTORE_EIP1283_CLEAR_REFUND = 15000
GAS_SSTORE_EIP1283_RESET_CLEAR_REFUND = 19800
GAS_SSTORE_EIP1283_RESET_REFUND = 4800

EIP1234_BLOCK_REWARD = 2 * denoms.ether



#
# New gas costs for some opcodes
#

# EIP-1108:
GAS_ECADD = 150
GAS_ECMUL = 6000
GAS_ECPAIRING_BASE = 45_000
GAS_ECPAIRING_PER_POINT = 34_000

# EIP-1884:
GAS_SLOAD_EIP1884 = 800
GAS_BALANCE_EIP1884 = 700
GAS_EXTCODEHASH_EIP1884 = 700

# New gas cost for transaction data
# EIP-2028
GAS_TXDATANONZERO_EIP2028 = 16



from eth.abc import (
    ComputationAPI,
)
from eth.exceptions import (
    InvalidInstruction,
)
from eth.vm.opcode import (
    Opcode,
)

INVALID_OPCODE_MAGIC_VAL = -0x1

class InvalidOpcode(Opcode):
    mnemonic = "INVALID"
    gas_cost = 0

    def __init__(self, value: int) -> None:
        self.value = value
        super().__init__()

    def __call__(self, computation: ComputationAPI) -> None:
        raise InvalidInstruction(
            f"Invalid opcode 0x{self.value:x} @ {computation.code.program_counter - 1}"
        )



from typing import (
    Tuple,
)

from eth_utils import (
    ValidationError,
    big_endian_to_int,
    int_to_big_endian,
)
from eth_utils.toolz import (
    curry,
)
from py_ecc import (
    optimized_bn128 as bn128,
)

from eth import (
    constants,
)
from eth._utils.bn128 import (
    validate_point,
)
from eth._utils.padding import (
    pad32,
    pad32r,
)
from eth.abc import (
    ComputationAPI,
)
from eth.exceptions import (
    VMError,
)


@curry
def ecmul(
    computation: ComputationAPI, gas_cost: int = constants.GAS_ECMUL
) -> ComputationAPI:
    computation.consume_gas(gas_cost, reason="ECMUL Precompile")

    try:
        result = _ecmull(computation.msg.data_as_bytes)
    except ValidationError:
        raise VMError("Invalid ECMUL parameters")

    result_x, result_y = result
    result_bytes = b"".join(
        (
            pad32(int_to_big_endian(result_x.n)),
            pad32(int_to_big_endian(result_y.n)),
        )
    )
    computation.output = result_bytes
    return computation


def _ecmull(data: bytes) -> Tuple[bn128.FQ, bn128.FQ]:
    x_bytes = pad32r(data[:32])
    y_bytes = pad32r(data[32:64])
    m_bytes = pad32r(data[64:96])

    x = big_endian_to_int(x_bytes)
    y = big_endian_to_int(y_bytes)
    m = big_endian_to_int(m_bytes)

    p = validate_point(x, y)

    result = bn128.normalize(bn128.multiply(p, m))
    return result



from functools import (
    partial,
)

from eth.vm.logic.storage import (
    NetSStoreGasSchedule,
    net_sstore,
)

GAS_SCHEDULE_EIP1283 = NetSStoreGasSchedule(
    sload_gas=200,
    sstore_set_gas=20000,
    sstore_reset_gas=5000,
    sstore_clears_schedule=15000,
)


sstore_eip1283 = partial(net_sstore, GAS_SCHEDULE_EIP1283)



from eth_utils import (
    ValidationError,
)

from eth.abc import (
    SignedTransactionAPI,
    StateAPI,
)
from eth.constants import (
    SECPK1_N,
)
from eth.vm.forks.frontier.validation import (
    validate_frontier_transaction,
)


def validate_homestead_transaction(
    state: StateAPI, transaction: SignedTransactionAPI
) -> None:
    if transaction.s > SECPK1_N // 2 or transaction.s == 0:
        raise ValidationError("Invalid signature S value")

    validate_frontier_transaction(state, transaction)



from eth_typing import (
    BlockNumber,
)

from eth.abc import (
    ComputationAPI,
)


def blockhash(computation: ComputationAPI) -> None:
    block_number = computation.stack_pop1_int()

    block_hash = computation.state.get_ancestor_hash(BlockNumber(block_number))

    computation.stack_push_bytes(block_hash)


def coinbase(computation: ComputationAPI) -> None:
    computation.stack_push_bytes(computation.state.coinbase)


def timestamp(computation: ComputationAPI) -> None:
    computation.stack_push_int(computation.state.timestamp)


def number(computation: ComputationAPI) -> None:
    computation.stack_push_int(computation.state.block_number)


def difficulty(computation: ComputationAPI) -> None:
    computation.stack_push_int(computation.state.difficulty)


def gaslimit(computation: ComputationAPI) -> None:
    computation.stack_push_int(computation.state.gas_limit)


def basefee(computation: ComputationAPI) -> None:
    computation.stack_push_int(computation.state.base_fee)


def mixhash(computation: ComputationAPI) -> None:
    computation.stack_push_bytes(computation.state.mix_hash)



import sys

from eth import (
    __version__,
)


def construct_evm_runtime_identifier() -> str:
    """
    Constructs the EVM runtime identifier string

    e.g. 'Py-EVM/v1.2.3/darwin-amd64/python3.9.13'
    """

    platform = sys.platform
    v = sys.version_info
    imp = sys.implementation

    return f"Py-EVM/{__version__}/{platform}/{imp.name}{v.major}.{v.minor}.{v.micro}"



from typing import (
    Tuple,
)

from eth_typing import (
    Address,
)
import rlp

from eth.consensus.clique.datatypes import (
    Snapshot,
    Tally,
    Vote,
    VoteAction,
)
from eth.rlp.sedes import (
    uint256,
)

ADDRESS_TALLY_SEDES = rlp.sedes.List((rlp.sedes.binary, rlp.sedes.binary))
VOTE_SEDES = rlp.sedes.List(
    (
        rlp.sedes.binary,
        uint256,
        rlp.sedes.binary,
        rlp.sedes.binary,
    )
)
SNAPSHOT_SEDES = rlp.sedes.List(
    (
        rlp.sedes.binary,
        rlp.sedes.CountableList(rlp.sedes.binary),
        rlp.sedes.CountableList(rlp.sedes.binary),
        rlp.sedes.CountableList(rlp.sedes.binary),
    )
)
TALLY_SEDES = rlp.sedes.List((rlp.sedes.binary, uint256))


def encode_address_tally_pair(pair: Tuple[Address, Tally]) -> bytes:
    return rlp.encode(
        [pair[0], encode_tally(pair[1])],
        sedes=ADDRESS_TALLY_SEDES,
    )


def decode_address_tally_pair(pair: bytes) -> Tuple[Address, Tally]:
    (
        address,
        tally_bytes,
    ) = rlp.decode(
        pair,
        sedes=ADDRESS_TALLY_SEDES,
    )

    tally = decode_tally(tally_bytes)

    return address, tally


def encode_vote(vote: Vote) -> bytes:
    return rlp.encode(
        [
            vote.signer,
            vote.block_number,
            vote.subject,
            vote.action.value,
        ],
        sedes=VOTE_SEDES,
    )


def decode_vote(vote: bytes) -> Vote:
    signer, block_number, subject, action = rlp.decode(
        vote,
        sedes=VOTE_SEDES,
    )
    return Vote(
        signer=signer,
        block_number=block_number,
        subject=subject,
        action=VoteAction.NOMINATE
        if action == VoteAction.NOMINATE.value
        else VoteAction.KICK,
    )


def encode_snapshot(snapshot: Snapshot) -> bytes:
    return rlp.encode(
        [
            snapshot.block_hash,
            list(snapshot.signers),
            [encode_vote(vote) for vote in snapshot.votes],
            [
                encode_address_tally_pair((address, tally))
                for address, tally in snapshot.tallies.items()
            ],
        ],
        sedes=SNAPSHOT_SEDES,
    )


def decode_snapshot(snapshot: bytes) -> Snapshot:
    block_hash, signers, votes_rlp, tallies_rlp = rlp.decode(
        snapshot,
        sedes=SNAPSHOT_SEDES,
    )

    votes = [decode_vote(vote) for vote in votes_rlp]
    tallies = dict(decode_address_tally_pair(pair) for pair in tallies_rlp)

    return Snapshot(
        signers=frozenset(signers),
        block_hash=block_hash,
        votes=frozenset(votes),
        tallies=tallies,
    )


def encode_tally(tally: Tally) -> bytes:
    return rlp.encode(
        [tally.action.value, tally.votes],
        sedes=TALLY_SEDES,
    )


def decode_tally(tally: bytes) -> Tally:
    action_binary, votes = rlp.decode(
        tally,
        sedes=TALLY_SEDES,
    )

    return Tally(action=VoteAction(action_binary), votes=votes)



import functools
from typing import (
    Any,
    Callable,
    TypeVar,
    cast,
)
import warnings

TFunc = TypeVar("TFunc", bound=Callable[..., Any])


def deprecate_method(func: TFunc, message: str = None) -> TFunc:
    @functools.wraps(func)
    def deprecated_func(*args: Any, **kwargs: Any) -> Any:
        warnings.warn(
            category=DeprecationWarning,
            message=(
                message
                or f"{func.__name__} is deprecated. "
                "A breaking change is expected in a future release."
            ),
            stacklevel=2,
        )
        func(*args, **kwargs)

    return cast(TFunc, deprecated_func)



import pytest

from eth_utils import (
    ValidationError,
)

from eth.abc import (
    AtomicDatabaseAPI,
)


class AtomicDatabaseBatchAPITestSuite:
    def test_atomic_batch_set_and_get(self, atomic_db: AtomicDatabaseAPI) -> None:
        with atomic_db.atomic_batch() as batch:
            batch.set(b"1", b"2")
            assert batch.get(b"1") == b"2"

        assert atomic_db.get(b"1") == b"2"

    def test_atomic_db_cannot_recursively_batch(
        self, atomic_db: AtomicDatabaseAPI
    ) -> None:
        with atomic_db.atomic_batch() as batch:
            assert not hasattr(batch, "atomic_batch")

    def test_atomic_db_with_set_and_delete_batch(
        self, atomic_db: AtomicDatabaseAPI
    ) -> None:
        atomic_db[b"key-1"] = b"origin"

        with atomic_db.atomic_batch() as batch:
            batch.delete(b"key-1")

            assert b"key-1" not in batch
            with pytest.raises(KeyError):
                assert batch[b"key-1"]

        with pytest.raises(KeyError):
            atomic_db[b"key-1"]

    def test_atomic_db_unbatched_sets_are_immediate(
        self, atomic_db: AtomicDatabaseAPI
    ) -> None:
        atomic_db[b"1"] = b"A"

        with atomic_db.atomic_batch() as batch:
            # Unbatched changes are immediate, and show up in batch reads
            atomic_db[b"1"] = b"B"
            assert batch[b"1"] == b"B"

            batch[b"1"] = b"C1"

            # It doesn't matter what changes happen underlying, all reads now
            # show the write applied to the batch db handle
            atomic_db[b"1"] = b"C2"
            assert batch[b"1"] == b"C1"

        # the batch write should overwrite any intermediate changes
        assert atomic_db[b"1"] == b"C1"

    def test_atomic_db_unbatched_deletes_are_immediate(
        self, atomic_db: AtomicDatabaseAPI
    ) -> None:
        atomic_db[b"1"] = b"A"

        with atomic_db.atomic_batch() as batch:
            assert b"1" in batch

            # Unbatched changes are immediate, and show up in batch reads
            del atomic_db[b"1"]

            assert b"1" not in batch

            batch[b"1"] = b"C1"

            # It doesn't matter what changes happen underlying, all reads now
            # show the write applied to the batch db handle
            atomic_db[b"1"] = b"C2"
            assert batch[b"1"] == b"C1"

        # the batch write should overwrite any intermediate changes
        assert atomic_db[b"1"] == b"C1"

    def test_atomic_db_cannot_use_batch_after_context(
        self, atomic_db: AtomicDatabaseAPI
    ) -> None:
        atomic_db[b"1"] = b"A"

        with atomic_db.atomic_batch() as batch:
            batch[b"1"] = b"B"

        # set
        with pytest.raises(ValidationError):
            batch[b"1"] = b"C"

        with pytest.raises(ValidationError):
            batch.set(b"1", b"C")

        # get
        with pytest.raises(ValidationError):
            batch[b"1"]

        with pytest.raises(ValidationError):
            batch.get(b"1")

        # exists
        with pytest.raises(ValidationError):
            assert b"1" in batch

        with pytest.raises(ValidationError):
            batch.exists(b"1")

        # delete
        with pytest.raises(ValidationError):
            del batch[b"1"]

        with pytest.raises(ValidationError):
            batch.delete(b"1")

        # none of the invalid changes above should change the original db
        assert atomic_db[b"1"] == b"B"

    def test_atomic_db_with_reverted_delete_batch(
        self, atomic_db: AtomicDatabaseAPI
    ) -> None:
        class CustomException(Exception):
            pass

        atomic_db[b"key-1"] = b"origin"

        with pytest.raises(CustomException):
            with atomic_db.atomic_batch() as batch:
                batch.delete(b"key-1")

                assert b"key-1" not in batch
                with pytest.raises(KeyError):
                    assert batch[b"key-1"]

                raise CustomException("pretend something went wrong")

        assert atomic_db[b"key-1"] == b"origin"

    def test_atomic_db_temporary_state_dropped_across_batches(
        self, atomic_db: AtomicDatabaseAPI
    ) -> None:
        class CustomException(Exception):
            pass

        atomic_db[b"key-1"] = b"origin"

        with pytest.raises(CustomException):
            with atomic_db.atomic_batch() as batch:
                batch.delete(b"key-1")
                batch.set(b"key-2", b"val-2")
                raise CustomException("pretend something went wrong")

        with atomic_db.atomic_batch() as batch:
            assert batch[b"key-1"] == b"origin"
            assert b"key-2" not in batch

    def test_atomic_db_with_exception_batch(self, atomic_db: AtomicDatabaseAPI) -> None:
        atomic_db.set(b"key-1", b"value-1")

        try:
            with atomic_db.atomic_batch() as batch:
                batch.set(b"key-1", b"new-value-1")
                batch.set(b"key-2", b"value-2")
                raise Exception
        except Exception:
            pass

        assert atomic_db.get(b"key-1") == b"value-1"

        with pytest.raises(KeyError):
            atomic_db[b"key-2"]



from typing import (
    List,
    Optional,
    Type,
    cast,
)

from eth_hash.auto import (
    keccak,
)
from eth_typing import (
    BlockNumber,
)
from eth_typing.evm import (
    Address,
    Hash32,
)
from eth_utils import (
    encode_hex,
)
import rlp
from rlp.sedes import (
    Binary,
    CountableList,
    big_endian_int,
    binary,
)

from eth._utils.headers import (
    new_timestamp_from_parent,
)
from eth.abc import (
    BlockHeaderAPI,
    BlockHeaderSedesAPI,
    MiningHeaderAPI,
    ReceiptBuilderAPI,
    TransactionBuilderAPI,
)
from eth.constants import (
    BLANK_ROOT_HASH,
    EMPTY_UNCLE_HASH,
    GENESIS_NONCE,
    GENESIS_PARENT_HASH,
    ZERO_ADDRESS,
    ZERO_HASH32,
)
from eth.rlp.headers import (
    BlockHeader,
)
from eth.rlp.sedes import (
    address,
    hash32,
    trie_root,
    uint256,
)
from eth.vm.forks.berlin.blocks import (
    BerlinBlock,
)

from .receipts import (
    LondonReceiptBuilder,
)
from .transactions import (
    LondonTransactionBuilder,
)

UNMINED_LONDON_HEADER_FIELDS = [
    ("parent_hash", hash32),
    ("uncles_hash", hash32),
    ("coinbase", address),
    ("state_root", trie_root),
    ("transaction_root", trie_root),
    ("receipt_root", trie_root),
    ("bloom", uint256),
    ("difficulty", big_endian_int),
    ("block_number", big_endian_int),
    ("gas_limit", big_endian_int),
    ("gas_used", big_endian_int),
    ("timestamp", big_endian_int),
    ("extra_data", binary),
    ("base_fee_per_gas", big_endian_int),
]


class LondonMiningHeader(rlp.Serializable, MiningHeaderAPI):
    fields = UNMINED_LONDON_HEADER_FIELDS


class LondonBlockHeader(rlp.Serializable, BlockHeaderAPI):
    fields = (
        UNMINED_LONDON_HEADER_FIELDS[:-1]
        + [
            ("mix_hash", binary),
            ("nonce", Binary(8, allow_empty=True)),
        ]
        + UNMINED_LONDON_HEADER_FIELDS[-1:]
    )

    def __init__(
        self,
        difficulty: int,
        block_number: BlockNumber,
        gas_limit: int,
        timestamp: int = None,
        coinbase: Address = ZERO_ADDRESS,
        parent_hash: Hash32 = ZERO_HASH32,
        uncles_hash: Hash32 = EMPTY_UNCLE_HASH,
        state_root: Hash32 = BLANK_ROOT_HASH,
        transaction_root: Hash32 = BLANK_ROOT_HASH,
        receipt_root: Hash32 = BLANK_ROOT_HASH,
        bloom: int = 0,
        gas_used: int = 0,
        extra_data: bytes = b"",
        mix_hash: Hash32 = ZERO_HASH32,
        nonce: bytes = GENESIS_NONCE,
        base_fee_per_gas: int = 0,
    ) -> None:
        if timestamp is None:
            if parent_hash == ZERO_HASH32:
                timestamp = new_timestamp_from_parent(None)
            else:
                # without access to the parent header,
                # we cannot select a new timestamp correctly
                raise ValueError(
                    "Must set timestamp explicitly if this is not a genesis header"
                )
        super().__init__(
            parent_hash=parent_hash,
            uncles_hash=uncles_hash,
            coinbase=coinbase,
            state_root=state_root,
            transaction_root=transaction_root,
            receipt_root=receipt_root,
            bloom=bloom,
            difficulty=difficulty,
            block_number=block_number,
            gas_limit=gas_limit,
            gas_used=gas_used,
            timestamp=timestamp,
            extra_data=extra_data,
            mix_hash=mix_hash,
            nonce=nonce,
            base_fee_per_gas=base_fee_per_gas,
        )

    def __str__(self) -> str:
        return f"<LondonBlockHeader #{self.block_number} {encode_hex(self.hash)[2:10]}>"

    _hash = None

    @property
    def hash(self) -> Hash32:
        if self._hash is None:
            self._hash = keccak(rlp.encode(self))
        return cast(Hash32, self._hash)

    @property
    def mining_hash(self) -> Hash32:
        non_pow_fields = self[:-3] + self[-1:]
        result = keccak(rlp.encode(non_pow_fields, LondonMiningHeader))
        return cast(Hash32, result)

    @property
    def hex_hash(self) -> str:
        return encode_hex(self.hash)

    @property
    def is_genesis(self) -> bool:
        # if removing the block_number == 0 test, consider the validation consequences.
        # validate_header stops trying to check the current header against
        # a parent header. Can someone trick us into following a high difficulty header
        # with genesis parent hash?
        return self.parent_hash == GENESIS_PARENT_HASH and self.block_number == 0

    @property
    def withdrawals_root(self) -> Optional[Hash32]:
        raise AttributeError("Withdrawals root not available until Shanghai fork")


class LondonBackwardsHeader(BlockHeaderSedesAPI):
    """
    An rlp sedes class for block headers.

    It can serialize and deserialize *both* London and pre-London headers.
    """

    @classmethod
    def serialize(cls, obj: BlockHeaderAPI) -> List[bytes]:
        return obj.serialize(obj)

    @classmethod
    def deserialize(cls, encoded: List[bytes]) -> BlockHeaderAPI:
        num_fields = len(encoded)
        if num_fields == 16:
            return LondonBlockHeader.deserialize(encoded)
        elif num_fields == 15:
            return BlockHeader.deserialize(encoded)
        else:
            raise ValueError(
                "London & earlier can only handle headers of 15 or 16 fields. "
                f"Got {num_fields} in {encoded!r}"
            )


class LondonBlock(BerlinBlock):
    transaction_builder: Type[TransactionBuilderAPI] = LondonTransactionBuilder
    receipt_builder: Type[ReceiptBuilderAPI] = LondonReceiptBuilder
    fields = [
        ("header", LondonBlockHeader),
        ("transactions", CountableList(transaction_builder)),
        ("uncles", CountableList(LondonBackwardsHeader)),
    ]



from eth_typing import (
    BlockNumber,
)

# https://github.com/ethereum/EIPs/blob/master/EIPS/eip-155.md
ROPSTEN_CHAIN_ID = 3

# Fork Blocks listed in ascending order

#
# DAO Block
#
DAO_FORK_ROPSTEN_BLOCK = BlockNumber(0)


#
# Tangerine Whistle Block
#
TANGERINE_WHISTLE_ROPSTEN_BLOCK = BlockNumber(0)


#
# Homestead Block
#
HOMESTEAD_ROPSTEN_BLOCK = BlockNumber(0)


#
# Spurious Dragon Block
#
SPURIOUS_DRAGON_ROPSTEN_BLOCK = BlockNumber(10)


#
# Byzantium Block
#
BYZANTIUM_ROPSTEN_BLOCK = BlockNumber(1700000)


#
# Constantinople
#
CONSTANTINOPLE_ROPSTEN_BLOCK = BlockNumber(4230000)


#
# Petersburg
#
PETERSBURG_ROPSTEN_BLOCK = BlockNumber(4939394)


#
# Istanbul
#
ISTANBUL_ROPSTEN_BLOCK = BlockNumber(6485846)

#
# Muir Glacier Block
#
MUIR_GLACIER_ROPSTEN_BLOCK = BlockNumber(7117117)
BERLIN_ROPSTEN_BLOCK = BlockNumber(9812189)



from typing import (
    Type,
)

from eth.abc import (
    TransactionExecutorAPI,
)

from ..arrow_glacier import (
    ArrowGlacierState,
)
from ..arrow_glacier.state import (
    ArrowGlacierTransactionExecutor,
)
from .computation import (
    GrayGlacierComputation,
)


class GrayGlacierTransactionExecutor(ArrowGlacierTransactionExecutor):
    pass


class GrayGlacierState(ArrowGlacierState):
    computation_class = GrayGlacierComputation
    transaction_executor_class: Type[
        TransactionExecutorAPI
    ] = GrayGlacierTransactionExecutor



from eth_typing import (
    Address,
)
from eth_utils import (
    encode_hex,
)

from eth import (
    constants,
)
from eth._utils.address import (
    force_bytes_to_address,
    generate_contract_address,
    generate_safe_contract_address,
)
from eth._utils.numeric import (
    ceil32,
)
from eth.abc import (
    ComputationAPI,
    MessageAPI,
)
from eth.exceptions import (
    Halt,
    Revert,
    WriteProtection,
)
from eth.vm import (
    mnemonics,
)
from eth.vm.opcode import (
    Opcode,
)

from .call import (
    max_child_gas_eip150,
)


def return_op(computation: ComputationAPI) -> None:
    start_position, size = computation.stack_pop_ints(2)

    computation.extend_memory(start_position, size)

    computation.output = computation.memory_read_bytes(start_position, size)
    raise Halt("RETURN")


def revert(computation: ComputationAPI) -> None:
    start_position, size = computation.stack_pop_ints(2)

    computation.extend_memory(start_position, size)

    computation.output = computation.memory_read_bytes(start_position, size)
    raise Revert(computation.output)


def selfdestruct(computation: ComputationAPI) -> None:
    beneficiary = force_bytes_to_address(computation.stack_pop1_bytes())
    _selfdestruct(computation, beneficiary)


def selfdestruct_eip150(computation: ComputationAPI) -> None:
    beneficiary = force_bytes_to_address(computation.stack_pop1_bytes())
    if not computation.state.account_exists(beneficiary):
        computation.consume_gas(
            constants.GAS_SELFDESTRUCT_NEWACCOUNT,
            reason=mnemonics.SELFDESTRUCT,
        )
    _selfdestruct(computation, beneficiary)


def selfdestruct_eip161_on_address(
    computation: ComputationAPI,
    beneficiary: Address,
) -> None:
    is_dead = not computation.state.account_exists(
        beneficiary
    ) or computation.state.account_is_empty(beneficiary)
    if is_dead and computation.state.get_balance(computation.msg.storage_address):
        computation.consume_gas(
            constants.GAS_SELFDESTRUCT_NEWACCOUNT,
            reason=mnemonics.SELFDESTRUCT,
        )
    _selfdestruct(computation, beneficiary)


def selfdestruct_eip161(computation: ComputationAPI) -> None:
    beneficiary = force_bytes_to_address(computation.stack_pop1_bytes())
    selfdestruct_eip161_on_address(computation, beneficiary)


def _selfdestruct(computation: ComputationAPI, beneficiary: Address) -> None:
    local_balance = computation.state.get_balance(computation.msg.storage_address)
    beneficiary_balance = computation.state.get_balance(beneficiary)

    # 1st: Transfer to beneficiary
    computation.state.set_balance(beneficiary, local_balance + beneficiary_balance)
    # 2nd: Zero the balance of the address being deleted (must come after
    # sending to beneficiary in case the contract named itself as the
    # beneficiary.
    computation.state.set_balance(computation.msg.storage_address, 0)

    computation.logger.debug2(
        "SELFDESTRUCT: %s (%s) -> %s",
        encode_hex(computation.msg.storage_address),
        local_balance,
        encode_hex(beneficiary),
    )

    # 3rd: Register the account to be deleted
    computation.register_account_for_deletion(beneficiary)
    raise Halt("SELFDESTRUCT")


class CreateOpcodeStackData:
    def __init__(
        self, endowment: int, memory_start: int, memory_length: int, salt: int = None
    ) -> None:
        self.endowment = endowment
        self.memory_start = memory_start
        self.memory_length = memory_length
        self.salt = salt


class Create(Opcode):
    def max_child_gas_modifier(self, gas: int) -> int:
        return gas

    def get_gas_cost(self, data: CreateOpcodeStackData) -> int:
        return self.gas_cost

    def generate_contract_address(
        self,
        stack_data: CreateOpcodeStackData,
        call_data: bytes,
        computation: ComputationAPI,
    ) -> Address:
        creation_nonce = computation.state.get_nonce(computation.msg.storage_address)
        computation.state.increment_nonce(computation.msg.storage_address)

        contract_address = generate_contract_address(
            computation.msg.storage_address,
            creation_nonce,
        )

        return contract_address

    def get_stack_data(self, computation: ComputationAPI) -> CreateOpcodeStackData:
        endowment, memory_start, memory_length = computation.stack_pop_ints(3)

        return CreateOpcodeStackData(endowment, memory_start, memory_length)

    def __call__(self, computation: ComputationAPI) -> None:
        stack_data = self.get_stack_data(computation)

        gas_cost = self.get_gas_cost(stack_data)
        computation.consume_gas(gas_cost, reason=self.mnemonic)

        computation.extend_memory(stack_data.memory_start, stack_data.memory_length)

        storage_address = computation.msg.storage_address
        storage_address_balance = computation.state.get_balance(storage_address)
        storage_address_nonce = computation.state.get_nonce(storage_address)

        insufficient_funds = storage_address_balance < stack_data.endowment
        stack_too_deep = computation.msg.depth + 1 > constants.STACK_DEPTH_LIMIT
        nonce_too_high = storage_address_nonce + 1 > constants.UINT_64_MAX

        if insufficient_funds or stack_too_deep or nonce_too_high:
            computation.stack_push_int(0)
            computation.return_data = b""
            if insufficient_funds:
                self.logger.debug2(
                    f"{self.mnemonic} failure: Insufficient Funds: "
                    f"{storage_address_balance} < {stack_data.endowment}",
                )
            elif stack_too_deep:
                self.logger.debug2(f"{self.mnemonic} failure: Stack limit reached")
            elif nonce_too_high:
                self.logger.debug2(f"{self.mnemonic} failure: Nonce too high")
            else:
                raise RuntimeError(
                    "Invariant: error must be insufficient funds or stack too deep"
                )
            return

        call_data = computation.memory_read_bytes(
            stack_data.memory_start, stack_data.memory_length
        )

        create_msg_gas = self.max_child_gas_modifier(computation.get_gas_remaining())
        computation.consume_gas(create_msg_gas, reason=self.mnemonic)

        contract_address = self.generate_contract_address(
            stack_data, call_data, computation
        )

        is_collision = computation.state.has_code_or_nonce(contract_address)

        if is_collision:
            computation.stack_push_int(0)
            computation.return_data = b""
            self.logger.debug2(
                "Address collision while creating contract: %s",
                encode_hex(contract_address),
            )
            return

        child_msg = computation.prepare_child_message(
            gas=create_msg_gas,
            to=constants.CREATE_CONTRACT_ADDRESS,
            value=stack_data.endowment,
            data=b"",
            code=call_data,
            create_address=contract_address,
        )
        self.apply_create_message(computation, child_msg)

    def apply_create_message(
        self,
        computation: ComputationAPI,
        child_msg: MessageAPI,
    ) -> None:
        child_computation = computation.apply_child_computation(child_msg)

        if child_computation.is_error:
            computation.stack_push_int(0)
        else:
            computation.stack_push_bytes(child_msg.storage_address)

        computation.return_gas(child_computation.get_gas_remaining())


class CreateEIP150(Create):
    def max_child_gas_modifier(self, gas: int) -> int:
        return max_child_gas_eip150(gas)


class CreateByzantium(CreateEIP150):
    def __call__(self, computation: ComputationAPI) -> None:
        if computation.msg.is_static:
            raise WriteProtection(
                "Cannot modify state while inside of a STATICCALL context"
            )
        return super().__call__(computation)


class Create2(CreateByzantium):
    def get_stack_data(self, computation: ComputationAPI) -> CreateOpcodeStackData:
        endowment, memory_start, memory_length, salt = computation.stack_pop_ints(4)

        return CreateOpcodeStackData(endowment, memory_start, memory_length, salt)

    def get_gas_cost(self, data: CreateOpcodeStackData) -> int:
        return (
            constants.GAS_CREATE
            + constants.GAS_SHA3WORD * ceil32(data.memory_length) // 32
        )

    def generate_contract_address(
        self,
        stack_data: CreateOpcodeStackData,
        call_data: bytes,
        computation: ComputationAPI,
    ) -> Address:
        computation.state.increment_nonce(computation.msg.storage_address)
        return generate_safe_contract_address(
            computation.msg.storage_address, stack_data.salt, call_data
        )

    def apply_create_message(
        self,
        computation: ComputationAPI,
        child_msg: MessageAPI,
    ) -> None:
        # We need to ensure that creation operates on empty storage **and**
        # that if the initialization code fails that we revert the account back
        # to its original state root.
        snapshot = computation.state.snapshot()

        computation.state.delete_storage(child_msg.storage_address)

        child_computation = computation.apply_child_computation(child_msg)

        if child_computation.is_error:
            computation.state.revert(snapshot)
            computation.stack_push_int(0)
        else:
            computation.state.commit(snapshot)
            computation.stack_push_bytes(child_msg.storage_address)

        computation.return_gas(child_computation.get_gas_remaining())



from typing import (
    Optional,
    cast,
)

from eth_utils.toolz import (
    curry,
)

from eth.abc import (
    SignedTransactionAPI,
    StateAPI,
)
from eth.exceptions import (
    VMError,
)
from eth.vm.spoof import (
    SpoofTransaction,
)


def _get_computation_error(
    state: StateAPI, transaction: SignedTransactionAPI
) -> Optional[VMError]:
    snapshot = state.snapshot()

    try:
        computation = state.apply_transaction(transaction)
        if computation.is_error:
            return computation.error
        else:
            return None

    finally:
        state.revert(snapshot)


@curry
def binary_gas_search(
    state: StateAPI, transaction: SignedTransactionAPI, tolerance: int = 1
) -> int:
    """
    Run the transaction with various gas limits, progressively
    approaching the minimum needed to succeed without an OutOfGas exception.

    The starting range of possible estimates is:
    [transaction.intrinsic_gas, state.gas_limit].
    After the first OutOfGas exception, the range is:
        (largest_limit_out_of_gas, state.gas_limit].
    After the first run not out of gas, the range is:
        (largest_limit_out_of_gas, smallest_success].

    :param int tolerance: When the range of estimates is less than tolerance,
        return the top of the range.
    :returns int: The smallest confirmed gas to not throw an OutOfGas exception,
        subject to tolerance. If OutOfGas is thrown at block limit, return block limit.
    :raises VMError: if the computation fails even when given the block gas_limit to
        complete
    """
    if not hasattr(transaction, "sender"):
        raise TypeError(
            "Transaction is missing attribute sender.",
            "If sending an unsigned transaction, use SpoofTransaction and provide the",
            "sender using the 'from' parameter",
        )

    minimum_transaction = cast(
        SignedTransactionAPI,
        SpoofTransaction(
            transaction,
            gas=transaction.intrinsic_gas,
            gas_price=0,
        ),
    )

    if _get_computation_error(state, minimum_transaction) is None:
        return transaction.intrinsic_gas

    maximum_transaction = cast(
        SignedTransactionAPI,
        SpoofTransaction(
            transaction,
            gas=state.gas_limit,
            gas_price=0,
        ),
    )
    error = _get_computation_error(state, maximum_transaction)
    if error is not None:
        raise error

    minimum_viable = state.gas_limit
    maximum_out_of_gas = transaction.intrinsic_gas
    while minimum_viable - maximum_out_of_gas > tolerance:
        midpoint = (minimum_viable + maximum_out_of_gas) // 2
        test_transaction = cast(
            SignedTransactionAPI, SpoofTransaction(transaction, gas=midpoint)
        )
        if _get_computation_error(state, test_transaction) is None:
            minimum_viable = midpoint
        else:
            maximum_out_of_gas = midpoint

    return minimum_viable


# Estimate in increments of intrinsic gas usage
binary_gas_search_intrinsic_tolerance = binary_gas_search(tolerance=21000)

# Estimate in increments of 1000 gas, takes roughly 5 more executions than
# intrinsic to estimate
binary_gas_search_1000_tolerance = binary_gas_search(tolerance=1000)

# Estimate to the exact gas, takes roughly 15 more executions than intrinsic to estimate
binary_gas_search_exact = binary_gas_search(tolerance=1)



from typing import (
    Iterable,
    Type,
)

from eth_utils import (
    to_tuple,
)

from eth.abc import (
    ConsensusAPI,
    VirtualMachineModifierAPI,
    VMConfiguration,
)
from eth.typing import (
    VMFork,
)


class ConsensusApplier(VirtualMachineModifierAPI):
    """
    This class is used to apply simple types of consensus engines to a series of
    virtual machines.

    Note that this *must not* be used for Clique, which has its own modifier
    """

    def __init__(self, consensus_class: Type[ConsensusAPI]) -> None:
        self._consensus_class = consensus_class

    @to_tuple
    def amend_vm_configuration(self, config: VMConfiguration) -> Iterable[VMFork]:
        """
        Amend the given ``VMConfiguration`` to operate
        under the rules of the pre-defined consensus
        """
        for pair in config:
            block_number, vm = pair
            yield block_number, vm.configure(consensus_class=self._consensus_class)



from eth_utils.toolz import (
    merge,
)

from eth import (
    precompiles,
)
from eth._utils.address import (
    force_bytes_to_address,
)
from eth.vm.forks.frontier.computation import (
    FRONTIER_PRECOMPILES,
)
from eth.vm.forks.spurious_dragon.computation import (
    SpuriousDragonComputation,
)

from .opcodes import (
    BYZANTIUM_OPCODES,
)

BYZANTIUM_PRECOMPILES = merge(
    FRONTIER_PRECOMPILES,
    {
        force_bytes_to_address(b"\x05"): precompiles.modexp,
        force_bytes_to_address(b"\x06"): precompiles.ecadd,
        force_bytes_to_address(b"\x07"): precompiles.ecmul,
        force_bytes_to_address(b"\x08"): precompiles.ecpairing,
    },
)


class ByzantiumComputation(SpuriousDragonComputation):
    """
    A class for all execution *message* computations in the ``Byzantium`` fork.
    Inherits from
    :class:`~eth.vm.forks.spurious_dragon.computation.SpuriousDragonComputation`
    """

    # Override
    opcodes = BYZANTIUM_OPCODES
    _precompiles = BYZANTIUM_PRECOMPILES



from typing import (
    Dict,
    Iterable,
    Set,
    Tuple,
    cast,
)

from eth_hash.auto import (
    keccak,
)
from eth_typing import (
    Address,
    Hash32,
)
from eth_utils import (
    ValidationError,
    encode_hex,
    get_extended_debug_logger,
    int_to_big_endian,
    to_checksum_address,
    to_dict,
    to_tuple,
)
from lru import (
    LRU,
)
import rlp
from trie import (
    HexaryTrie,
    exceptions as trie_exceptions,
)

from eth.abc import (
    AccountDatabaseAPI,
    AccountStorageDatabaseAPI,
    AtomicDatabaseAPI,
    DatabaseAPI,
    MetaWitnessAPI,
)
from eth.constants import (
    BLANK_ROOT_HASH,
    EMPTY_SHA3,
)
from eth.db.accesslog import (
    KeyAccessLoggerAtomicDB,
    KeyAccessLoggerDB,
)
from eth.db.backends.memory import (
    MemoryDB,
)
from eth.db.batch import (
    BatchDB,
)
from eth.db.cache import (
    CacheDB,
)
from eth.db.diff import (
    DBDiff,
)
from eth.db.journal import (
    JournalDB,
)
from eth.db.storage import (
    AccountStorageDB,
)
from eth.db.witness import (
    AccountQueryTracker,
    MetaWitness,
)
from eth.rlp.accounts import (
    Account,
    LazyAccount,
)
from eth.typing import (
    JournalDBCheckpoint,
)
from eth.validation import (
    validate_canonical_address,
    validate_is_bytes,
    validate_uint64,
    validate_uint256,
)
from eth.vm.interrupt import (
    MissingAccountTrieNode,
    MissingBytecode,
)

from .hash_trie import (
    HashTrie,
)

from web3 import Web3
from web3.types import (
    BlockIdentifier,
)

IS_PRESENT_VALUE = b""


class AccountDB(AccountDatabaseAPI):
    logger = get_extended_debug_logger("eth.db.account.AccountDB")

    def __init__(
        self, w3: Web3, block_identifier: BlockIdentifier
    ) -> None:
        r"""
        Internal implementation details (subject to rapid change):
        Database entries go through several pipes, like so...

        .. code::

            db > _batchdb ------------------------> _journaldb -----------> code lookups
             \
              -> _batchtrie -> _trie -> _trie_cache -> _journaltrie ---> account lookups

        Journaling sequesters writes at the _journal* attrs ^, until persist is called.

        _batchtrie enables us to prune all trie changes while building
        state,  without deleting old trie roots.

        _batchdb and _batchtrie together enable us to make the state root,
        without saving everything to the database.

        _journaldb is a journaling of the keys and values used to store
        code and account storage.

        _trie is a hash-trie, used to generate the state root

        _trie_cache is a cache tied to the state root of the trie. It
        is important that this cache is checked *after* looking for
        the key in _journaltrie, because the cache is only invalidated
        after a state root change.

        _journaltrie is a journaling of the accounts (an address->rlp mapping,
        rather than the nodes stored by the trie). This enables
        a squashing of all account changes before pushing them into the trie.

        .. NOTE:: StorageDB works similarly

        AccountDB synchronizes the snapshot/revert/persist of both of the
        journals.
        """
        self.w3 = w3
        self.block_identifier = block_identifier
        self._journaldb = JournalDB(KeyAccessLoggerAtomicDB(MemoryDB()))
        self._account_cache = LRU(1024 * 10)
        self._account_stores: Dict[Address, AccountStorageDatabaseAPI] = {}
        self._dirty_accounts: Set[Address] = set()
        self._accessed_accounts: Set[Address] = set()
        self._accessed_bytecodes: Set[Address] = set()
        # Track whether an account or slot have been accessed
        # during a given transaction:
        self._reset_access_counters()

    @property
    def state_root(self) -> Hash32:
        raise NotImplementedError("too lazy to implement this")
        return self._trie.root_hash

    @state_root.setter
    def state_root(self, value: Hash32) -> None:
        raise NotImplementedError("too lazy to implement this")
        if self._trie.root_hash != value:
            self._trie_cache.reset_cache()
            self._trie.root_hash = value

    def has_root(self, state_root: bytes) -> bool:
        raise NotImplementedError("too lazy to implement this")
        return state_root in self._batchtrie

    #
    # Storage
    #
    def get_storage(
        self, address: Address, slot: int, from_journal: bool = True
    ) -> int:
        validate_canonical_address(address, title="Storage Address")
        validate_uint256(slot, title="Storage Slot")

        account_store = self._get_address_store(address)
        return account_store.get(slot, from_journal)

    def set_storage(self, address: Address, slot: int, value: int) -> None:
        validate_uint256(value, title="Storage Value")
        validate_uint256(slot, title="Storage Slot")
        validate_canonical_address(address, title="Storage Address")

        account_store = self._get_address_store(address)
        self._dirty_accounts.add(address)
        account_store.set(slot, value)

    def delete_storage(self, address: Address) -> None:
        validate_canonical_address(address, title="Storage Address")

        self._set_storage_root(address, BLANK_ROOT_HASH)
        self._wipe_storage(address)

    def is_storage_warm(self, address: Address, slot: int) -> bool:
        key = self._get_storage_tracker_key(address, slot)
        return key in self._journal_accessed_state

    def mark_storage_warm(self, address: Address, slot: int) -> None:
        key = self._get_storage_tracker_key(address, slot)
        if key not in self._journal_accessed_state:
            self._journal_accessed_state[key] = IS_PRESENT_VALUE

    def _get_storage_tracker_key(self, address: Address, slot: int) -> bytes:
        """
        Get the key used to track whether a storage slot has been accessed
        during this transaction.
        """
        return address + int_to_big_endian(slot)

    def _wipe_storage(self, address: Address) -> None:
        """
        Wipe out the storage, without explicitly handling the storage root update
        """
        account_store = self._get_address_store(address)
        self._dirty_accounts.add(address)
        account_store.delete()

    def _get_address_store(self, address: Address) -> AccountStorageDatabaseAPI:
        if address in self._account_stores:
            store = self._account_stores[address]
        else:
            store = AccountStorageDB({}, self.w3, self.block_identifier, address)
            self._account_stores[address] = store
        return store

    def _dirty_account_stores(
        self,
    ) -> Iterable[Tuple[Address, AccountStorageDatabaseAPI]]:
        for address in self._dirty_accounts:
            store = self._account_stores[address]
            yield address, store

    @to_tuple
    def _get_changed_roots(self) -> Iterable[Tuple[Address, Hash32]]:
        # list all the accounts that were changed, and their new storage roots
        for address, store in self._dirty_account_stores():
            if store.has_changed_root:
                yield address, store.get_changed_root()

    def _get_storage_root(self, address: Address) -> Hash32:
        account = self._get_account(address)
        return account.storage_root

    def _set_storage_root(self, address: Address, new_storage_root: Hash32) -> None:
        account = self._get_account(address)
        self._set_account(address, account.copy(storage_root=new_storage_root))

    def _validate_flushed_storage(
        self, address: Address, store: AccountStorageDatabaseAPI
    ) -> None:
        if store.has_changed_root:
            actual_storage_root = self._get_storage_root(address)
            expected_storage_root = store.get_changed_root()
            if expected_storage_root != actual_storage_root:
                raise ValidationError(
                    "Storage root was not saved to account before trying to persist "
                    f"roots. Account {address!r} had storage {actual_storage_root!r}, "
                    f"but should be {expected_storage_root!r}."
                )

    #
    # Balance
    #
    def get_balance(self, address: Address) -> int:
        validate_canonical_address(address, title="Storage Address")

        account = self._get_account(address)
        return account.balance

    def set_balance(self, address: Address, balance: int) -> None:
        validate_canonical_address(address, title="Storage Address")
        validate_uint256(balance, title="Account Balance")

        account = self._get_account(address)
        self._set_account(address, account.copy(balance=balance))

    #
    # Nonce
    #
    def get_nonce(self, address: Address) -> int:
        validate_canonical_address(address, title="Storage Address")

        account = self._get_account(address)
        return account.nonce

    def set_nonce(self, address: Address, nonce: int) -> None:
        validate_canonical_address(address, title="Storage Address")
        validate_uint64(nonce, title="Nonce")

        account = self._get_account(address)
        self._set_account(address, account.copy(nonce=nonce))

    def increment_nonce(self, address: Address) -> None:
        current_nonce = self.get_nonce(address)
        self.set_nonce(address, current_nonce + 1)

    #
    # Code
    #
    def get_code(self, address: Address) -> bytes:
        validate_canonical_address(address, title="Storage Address")

        bkey = address + b"code"

        if bkey in self._journaldb:
            ret = self._journaldb[bkey]
        else:
            ret = self.w3.eth.get_code(address, block_identifier=self.block_identifier)
            self._journaldb[bkey] = ret
        
        if ret != b"":
            self._accessed_bytecodes.add(address)

        return ret

        # code_hash = self.get_code_hash(address)
        # if code_hash == EMPTY_SHA3:
        #     return b""
        # else:
        #     try:
        #         return self._journaldb[code_hash]
        #     except KeyError:
        #         raise MissingBytecode(code_hash) from KeyError
        #     finally:
        #         if code_hash in self._get_accessed_node_hashes():
        #             self._accessed_bytecodes.add(address)

    def set_code(self, address: Address, code: bytes) -> None:
        validate_canonical_address(address, title="Storage Address")
        validate_is_bytes(code, title="Code")

        # account = self._get_account(address)

        bkey = address + b"code"
        self._journaldb[bkey] = code
        # self._set_account(address, account.copy(code_hash=code_hash))

    def get_code_hash(self, address: Address) -> Hash32:
        validate_canonical_address(address, title="Storage Address")

        code = self.get_code(address)
        h = keccak(code)
        return h

    def delete_code(self, address: Address) -> None:
        validate_canonical_address(address, title="Storage Address")

        # set code to empty
        key = address + b"code"
        self._journaldb[key] = b""

        # account = self._get_account(address)
        # self._set_account(address, account.copy(code_hash=EMPTY_SHA3))

    #
    # Account Methods
    #
    def account_has_code_or_nonce(self, address: Address) -> bool:
        return self.get_nonce(address) != 0 or self.get_code(address) != b''

    def delete_account(self, address: Address) -> None:
        validate_canonical_address(address, title="Storage Address")

        # We must wipe the storage first, because if it's the first time we load it,
        #   then we want to load it with the original storage root hash, not the
        #   empty one. (in case of a later revert, we don't want to poison the
        #   storage cache)
        self._wipe_storage(address)

        if address in self._account_cache:
            del self._account_cache[address]

        # set code to empty
        key = address + b"code"
        self._journaldb[key] = b""

    def account_exists(self, address: Address) -> bool:
        validate_canonical_address(address, title="Storage Address")
        # NOTE: not sure if this works since we technically also 
        # need to see if the account's storage (storage_root) is empty, I think?
        return self.account_is_empty(address) is False

    def touch_account(self, address: Address) -> None:
        validate_canonical_address(address, title="Storage Address")

        account = self._get_account(address)
        self._set_account(address, account)

    def account_is_empty(self, address: Address) -> bool:
        return (
            not self.account_has_code_or_nonce(address)
            and self.get_balance(address) == 0
        )

    def is_address_warm(self, address: Address) -> bool:
        return address in self._journal_accessed_state

    def mark_address_warm(self, address: Address) -> None:
        if address not in self._journal_accessed_state:
            self._journal_accessed_state[address] = IS_PRESENT_VALUE

    #
    # Internal
    #
    def _get_encoded_account(
        self, address: Address, from_journal: bool = True
    ) -> bytes:
        self._accessed_accounts.add(address)
        lookup_trie = self._journaltrie if from_journal else self._trie_cache

        try:
            return lookup_trie[address]
        except trie_exceptions.MissingTrieNode as exc:
            raise MissingAccountTrieNode(*exc.args) from exc
        except KeyError:
            # In case the account is deleted in the JournalDB
            return b""

    @staticmethod
    def _make_account_journal_key(address: Address) -> bytes:
        return b'account:' + address

    def _get_account(self, address: Address, from_journal: bool = True) -> Account:
        if from_journal:
            if address in self._account_cache:
                return self._account_cache[address]
            if address in self._journaldb:
                return self._journaldb[address]

        account = LazyAccount(self.w3, self.block_identifier, address)

        if from_journal:
            self._account_cache[address] = account
        return account

    def _set_account(self, address: Address, account: Account) -> None:
        assert account is not None
        self._journaldb[address] = account
        self._account_cache[address] = account

    def _reset_access_counters(self) -> None:
        # Account accesses and storage accesses recorded in the same journal
        # Accounts just use the address as the key (and an empty value as a flag)
        # Storage use a concatenation of address and slot converted to bytes
        # (and empty value)
        self._journal_accessed_state = JournalDB(MemoryDB())

    #
    # Record and discard API
    #
    def record(self) -> JournalDBCheckpoint:
        checkpoint = self._journaldb.record()
        self._journal_accessed_state.record(checkpoint)

        for _, store in self._dirty_account_stores():
            store.record(checkpoint)
        return checkpoint

    def discard(self, checkpoint: JournalDBCheckpoint) -> None:
        self._journaldb.discard(checkpoint)
        try:
            self._journal_accessed_state.discard(checkpoint)
        except ValidationError as e:
            # ignore missing checkpoint errors
            if 'No checkpoint' not in str(e):
                raise
        self._account_cache.clear()
        for _, store in self._dirty_account_stores():
            store.discard(checkpoint)

    def commit(self, checkpoint: JournalDBCheckpoint) -> None:
        self._journaldb.commit(checkpoint)
        self._journal_accessed_state.commit(checkpoint)
        for _, store in self._dirty_account_stores():
            store.commit(checkpoint)

    def lock_changes(self) -> None:
        for _, store in self._dirty_account_stores():
            store.lock_changes()
        self._reset_access_counters()

    def make_state_root(self) -> Hash32:
        for _address, store in self._dirty_account_stores():
            store.make_storage_root()

        for address, storage_root in self._get_changed_roots():
            if self.account_exists(address) or storage_root != BLANK_ROOT_HASH:
                self.logger.debug2(
                    "Updating account 0x%s to storage root 0x%s",
                    address.hex(),
                    storage_root.hex(),
                )
                self._set_storage_root(address, storage_root)

        self._journaldb.persist()

        diff = self._journaltrie.diff()
        if diff.deleted_keys() or diff.pending_items():
            # In addition to squashing (which is redundant here), this context manager
            # causes an atomic commit of the changes, so exceptions will revert the trie
            with self._trie.squash_changes() as memory_trie:
                self._apply_account_diff_without_proof(diff, memory_trie)

        self._journaltrie.reset()
        self._trie_cache.reset_cache()

        return self.state_root

    def persist(self) -> MetaWitnessAPI:
        self.make_state_root()

        # persist storage
        with self._raw_store_db.atomic_batch() as write_batch:
            for address, store in self._dirty_account_stores():
                self._validate_flushed_storage(address, store)
                store.persist(write_batch)

        for address, new_root in self._get_changed_roots():
            if new_root is None:
                raise ValidationError(
                    f"Cannot validate new root of account 0x{address.hex()} "
                    f"which has a new root hash of None"
                )
            elif new_root not in self._raw_store_db and new_root != BLANK_ROOT_HASH:
                raise ValidationError(
                    "After persisting storage trie, a root node was not found. "
                    f"State root for account 0x{address.hex()} "
                    f"is missing for hash 0x{new_root.hex()}."
                )

        # generate witness (copy) before clearing the underlying data
        meta_witness = self._get_meta_witness()

        # reset local storage trackers
        self._account_stores = {}
        self._dirty_accounts = set()
        self._accessed_accounts = set()
        self._accessed_bytecodes = set()
        # We have to clear the account cache here so that future account accesses
        #   will get added to _accessed_accounts correctly. Account accesses that
        #   are cached do not add the address to the list of accessed accounts.
        self._account_cache.clear()

        # persist accounts
        self._validate_generated_root()
        new_root_hash = self.state_root
        self.logger.debug2("Persisting new state root: 0x%s", new_root_hash.hex())
        with self._raw_store_db.atomic_batch() as write_batch:
            self._batchtrie.commit_to(write_batch, apply_deletes=False)
            self._batchdb.commit_to(write_batch, apply_deletes=False)
        self._root_hash_at_last_persist = new_root_hash

        return meta_witness

    def _get_accessed_node_hashes(self) -> Set[Hash32]:
        return cast(Set[Hash32], self._raw_store_db.keys_read)

    @to_dict
    def _get_access_list(self) -> Iterable[Tuple[Address, AccountQueryTracker]]:
        """
        Get the list of addresses that were accessed, whether the bytecode was accessed,
        and which storage slots were accessed.
        """
        for address in self._accessed_accounts:
            did_access_bytecode = address in self._accessed_bytecodes
            if address in self._account_stores:
                accessed_storage_slots = self._account_stores[
                    address
                ].get_accessed_slots()
            else:
                accessed_storage_slots = frozenset()
            yield address, AccountQueryTracker(
                did_access_bytecode, accessed_storage_slots
            )

    def _get_meta_witness(self) -> MetaWitness:
        """
        Get a variety of metadata about the state witness needed to execute the block.

        This creates a copy, so that underlying changes do not affect the returned
        MetaWitness.
        """
        return MetaWitness(self._get_accessed_node_hashes(), self._get_access_list())

    def _validate_generated_root(self) -> None:
        db_diff = self._journaldb.diff()
        if db_diff:
            raise ValidationError(
                f"AccountDB had a dirty db when it needed to be clean: {db_diff!r}"
            )
        trie_diff = self._journaltrie.diff()
        if trie_diff:
            raise ValidationError(
                f"AccountDB had a dirty trie when it needed to be clean: {trie_diff!r}"
            )

    def _log_pending_accounts(self) -> None:
        # This entire method is about debug2 logging, so skip it if debug2 is disabled
        if not self.logger.show_debug2:
            return

        diff = self._journaltrie.diff()
        for address in sorted(diff.pending_keys()):
            account = self._get_account(Address(address))
            self.logger.debug2(
                "Pending Account %s: balance %d, nonce %d, storage root %s, code hash %s",  # noqa: E501
                to_checksum_address(address),
                account.balance,
                account.nonce,
                encode_hex(account.storage_root),
                encode_hex(account.code_hash),
            )
        for deleted_address in sorted(diff.deleted_keys()):
            # Check if the account was accessed before accessing/logging
            # info about the address
            was_account_accessed = deleted_address in self._accessed_accounts
            cast_deleted_address = Address(deleted_address)
            self.logger.debug2(
                "Deleted Account %s, empty? %s, exists? %s",
                to_checksum_address(deleted_address),
                self.account_is_empty(cast_deleted_address),
                self.account_exists(cast_deleted_address),
            )
            # If the account was not accessed previous to the log,
            # (re)mark it as not accessed
            if not was_account_accessed:
                self._accessed_accounts.remove(cast_deleted_address)

    def _apply_account_diff_without_proof(
        self, diff: DBDiff, trie: DatabaseAPI
    ) -> None:
        """
        Apply diff of trie updates, when original nodes might be missing.
        Note that doing this naively will raise exceptions about missing nodes
        from *intermediate* trie roots. This captures exceptions and uses the previous
        trie root hash that will be recognized by other nodes.
        """
        # It's fairly common that when an account is deleted, we need to retrieve nodes
        # for accounts that were not needed during normal execution. We only need these
        # nodes to refactor the trie.
        for delete_key in diff.deleted_keys():
            try:
                del trie[delete_key]
            except trie_exceptions.MissingTrieNode as exc:
                self.logger.debug(
                    "Missing node while deleting account with key %s: %s",
                    encode_hex(delete_key),
                    exc,
                )
                raise MissingAccountTrieNode(
                    exc.missing_node_hash,
                    self._root_hash_at_last_persist,
                    exc.requested_key,
                ) from exc

        # It's fairly unusual, but possible, that setting an account will need unknown
        # nodes during a trie refactor. Here is an example that seems to cause it:
        #
        # Setup:
        #   - Root node is a branch, with 0 pointing to a leaf
        #   - The complete leaf key is (0, 1, 2), so (1, 2) is in the leaf node
        #   - We know the leaf node hash but not the leaf node body
        # Refactor that triggers missing node:
        #   - Add value with key (0, 3, 4)
        #   - We need to replace the current leaf node with a branch that points leaves
        #       at 1 and 3
        #   - The leaf for key (0, 1, 2) now contains only the (2) part, so needs to
        #       be rebuilt
        #   - We need the full body of the old (1, 2) leaf node, to rebuild

        for key, val in diff.pending_items():
            try:
                trie[key] = val
            except trie_exceptions.MissingTrieNode as exc:
                self.logger.debug(
                    "Missing node on account update key %s to %s: %s",
                    encode_hex(key),
                    encode_hex(val),
                    exc,
                )
                raise MissingAccountTrieNode(
                    exc.missing_node_hash,
                    self._root_hash_at_last_persist,
                    exc.requested_key,
                ) from exc



from abc import (
    ABC,
)
from typing import (
    Type,
)

from eth_utils import (
    encode_hex,
)
from rlp.sedes import (
    CountableList,
)

from eth.abc import (
    TransactionBuilderAPI,
)
from eth.vm.forks.gray_glacier.blocks import (
    GrayGlacierBlock,
    GrayGlacierBlockHeader,
    GrayGlacierMiningHeader,
)

from ..london.blocks import (
    LondonBackwardsHeader,
)
from .transactions import (
    ParisTransactionBuilder,
)


class ParisMiningHeader(GrayGlacierMiningHeader, ABC):
    pass


class ParisBlockHeader(GrayGlacierBlockHeader, ABC):
    def __str__(self) -> str:
        return f"<ParisBlockHeader #{self.block_number} {encode_hex(self.hash)[2:10]}>"


class ParisBlock(GrayGlacierBlock):
    transaction_builder: Type[TransactionBuilderAPI] = ParisTransactionBuilder
    fields = [
        ("header", ParisBlockHeader),
        ("transactions", CountableList(transaction_builder)),
        # no uncles in pos, max_length=0
        ("uncles", CountableList(LondonBackwardsHeader, max_length=0)),
    ]



from typing import (
    Dict,
    FrozenSet,
    NamedTuple,
    Set,
)

from eth_typing import (
    Address,
    Hash32,
)

from eth.abc import (
    MetaWitnessAPI,
)


class AccountQueryTracker(NamedTuple):
    did_query_bytecode: bool
    slots_queried: FrozenSet[int]


class MetaWitness(MetaWitnessAPI):
    def __init__(
        self,
        witness_hashes: Set[Hash32],
        accounts_metadata_queried: Dict[Address, AccountQueryTracker],
    ) -> None:
        self._trie_node_hashes = frozenset(witness_hashes)
        self._accounts_metadata_queried = accounts_metadata_queried

    @property
    def hashes(self) -> FrozenSet[Hash32]:
        return self._trie_node_hashes

    @property
    def accounts_queried(self) -> FrozenSet[Address]:
        return frozenset(self._accounts_metadata_queried.keys())

    @property
    def account_bytecodes_queried(self) -> FrozenSet[Address]:
        return frozenset(
            address
            for address, query_tracker in self._accounts_metadata_queried.items()
            if query_tracker.did_query_bytecode
        )

    def get_slots_queried(self, address: Address) -> FrozenSet[int]:
        try:
            query_tracker = self._accounts_metadata_queried[address]
        except KeyError:
            return frozenset()
        else:
            return query_tracker.slots_queried

    @property
    def total_slots_queried(self) -> int:
        """
        Summed across all accounts, how many storage slots were queried?
        """
        return sum(
            len(query_tracker.slots_queried)
            for query_tracker in self._accounts_metadata_queried.values()
        )



import logging
import operator
import random
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Sequence,
    Tuple,
    Type,
)

from eth_typing import (
    Address,
    BlockNumber,
    Hash32,
)
from eth_utils import (
    encode_hex,
)
from eth_utils.toolz import (
    concatv,
    sliding_window,
)

from eth._utils.datatypes import (
    Configurable,
)
from eth._utils.db import (
    apply_state_dict,
)
from eth._utils.rlp import (
    validate_imported_block_unchanged,
)
from eth._warnings import (
    catch_and_ignore_import_warning,
)
from eth.abc import (
    AtomicDatabaseAPI,
    BlockAndMetaWitness,
    BlockAPI,
    BlockHeaderAPI,
    BlockImportResult,
    BlockPersistResult,
    ChainAPI,
    ChainDatabaseAPI,
    ComputationAPI,
    ConsensusContextAPI,
    MiningChainAPI,
    ReceiptAPI,
    SignedTransactionAPI,
    StateAPI,
    UnsignedTransactionAPI,
    VirtualMachineAPI,
    WithdrawalAPI,
)
from eth.consensus import (
    ConsensusContext,
)
from eth.constants import (
    EMPTY_UNCLE_HASH,
    MAX_UNCLE_DEPTH,
)
from eth.db.chain import (
    ChainDB,
)
from eth.db.header import (
    HeaderDB,
)
from eth.estimators import (
    get_gas_estimator,
)
from eth.exceptions import (
    HeaderNotFound,
    TransactionNotFound,
    VMNotFound,
)
from eth.rlp.headers import (
    BlockHeader,
)
from eth.typing import (
    AccountState,
    HeaderParams,
    StaticMethod,
)
from eth.validation import (
    validate_block_number,
    validate_uint256,
    validate_vm_configuration,
    validate_word,
)
from eth.vm.chain_context import (
    ChainContext,
)

with catch_and_ignore_import_warning():
    from eth_utils import (
        ValidationError,
        to_set,
    )
    from eth_utils.toolz import (
        assoc,
        compose,
        groupby,
        iterate,
        take,
    )


class BaseChain(Configurable, ChainAPI):
    """
    The base class for all Chain objects
    """

    chaindb: ChainDatabaseAPI = None
    chaindb_class: Type[ChainDatabaseAPI] = None
    consensus_context_class: Type[ConsensusContextAPI] = None
    vm_configuration: Tuple[Tuple[BlockNumber, Type[VirtualMachineAPI]], ...] = None
    chain_id: int = None

    @classmethod
    def get_vm_class_for_block_number(
        cls, block_number: BlockNumber
    ) -> Type[VirtualMachineAPI]:
        if cls.vm_configuration is None:
            raise AttributeError(
                "Chain classes must define the VMs in vm_configuration"
            )

        validate_block_number(block_number)
        for start_block, vm_class in reversed(cls.vm_configuration):
            if block_number >= start_block:
                return vm_class
        else:
            raise VMNotFound(f"No vm available for block #{block_number}")

    @classmethod
    def get_vm_class(cls, header: BlockHeaderAPI) -> Type[VirtualMachineAPI]:
        return cls.get_vm_class_for_block_number(header.block_number)

    #
    # Validation API
    #
    def validate_chain(
        self,
        root: BlockHeaderAPI,
        descendants: Tuple[BlockHeaderAPI, ...],
        seal_check_random_sample_rate: int = 1,
    ) -> None:
        all_indices = range(len(descendants))
        if seal_check_random_sample_rate == 1:
            indices_to_check_seal = set(all_indices)
        elif seal_check_random_sample_rate == 0:
            indices_to_check_seal = set()
        else:
            sample_size = len(all_indices) // seal_check_random_sample_rate
            indices_to_check_seal = set(random.sample(all_indices, sample_size))

        header_pairs = sliding_window(2, concatv([root], descendants))

        for index, (parent, child) in enumerate(header_pairs):
            if child.parent_hash != parent.hash:
                raise ValidationError(
                    f"Invalid header chain; {child} has parent"
                    f" {encode_hex(child.parent_hash)},"
                    f" but expected {encode_hex(parent.hash)}"
                )
            vm = self.get_vm(child)
            try:
                vm.validate_header(child, parent)
            except ValidationError as exc:
                raise ValidationError(
                    f"{child} is not a valid child of {parent}: {exc}"
                ) from exc

            if index in indices_to_check_seal:
                vm.validate_seal(child)

    def validate_chain_extension(self, headers: Tuple[BlockHeaderAPI, ...]) -> None:
        for index, header in enumerate(headers):
            vm = self.get_vm(header)

            # pass in any parents that are not already in the database
            parents = headers[:index]
            vm.validate_seal_extension(header, parents)


class Chain(BaseChain):
    logger = logging.getLogger("eth.chain.chain.Chain")
    gas_estimator: StaticMethod[Callable[[StateAPI, SignedTransactionAPI], int]] = None

    chaindb_class: Type[ChainDatabaseAPI] = ChainDB
    consensus_context_class: Type[ConsensusContextAPI] = ConsensusContext

    def __init__(self, base_db: AtomicDatabaseAPI) -> None:
        if not self.vm_configuration:
            raise ValueError(
                "The Chain class cannot be instantiated with "
                "an empty `vm_configuration`"
            )
        else:
            validate_vm_configuration(self.vm_configuration)

        self.chaindb = self.get_chaindb_class()(base_db)
        self.consensus_context = self.consensus_context_class(self.chaindb.db)
        self.headerdb = HeaderDB(base_db)
        if self.gas_estimator is None:
            self.gas_estimator = get_gas_estimator()

    #
    # Helpers
    #
    @classmethod
    def get_chaindb_class(cls) -> Type[ChainDatabaseAPI]:
        if cls.chaindb_class is None:
            raise AttributeError("`chaindb_class` not set")
        return cls.chaindb_class

    #
    # Chain API
    #
    @classmethod
    def from_genesis(
        cls,
        base_db: AtomicDatabaseAPI,
        genesis_params: Dict[str, HeaderParams],
        genesis_state: AccountState = None,
    ) -> "BaseChain":
        genesis_vm_class = cls.get_vm_class_for_block_number(BlockNumber(0))

        pre_genesis_header = BlockHeader(difficulty=0, block_number=-1, gas_limit=0)
        chain_context = ChainContext(cls.chain_id)
        state = genesis_vm_class.build_state(base_db, pre_genesis_header, chain_context)

        if genesis_state is None:
            genesis_state = {}

        # mutation
        apply_state_dict(state, genesis_state)
        state.persist()

        if "state_root" not in genesis_params:
            # If the genesis state_root was not specified, use the value
            # computed from the initialized state database.
            genesis_params = assoc(genesis_params, "state_root", state.state_root)
        elif genesis_params["state_root"] != state.state_root:
            # If the genesis state_root was specified, validate that it matches
            # the computed state from the initialized state database.
            raise ValidationError(
                "The provided genesis state root does not match the computed "
                f"genesis state root.  Got {state.state_root!r}.  "
                f"Expected {genesis_params['state_root']!r}"
            )

        genesis_header = genesis_vm_class.create_genesis_header(**genesis_params)
        return cls.from_genesis_header(base_db, genesis_header)

    @classmethod
    def from_genesis_header(
        cls, base_db: AtomicDatabaseAPI, genesis_header: BlockHeaderAPI
    ) -> "BaseChain":
        chaindb = cls.get_chaindb_class()(base_db)
        chaindb.persist_header(genesis_header)
        return cls(base_db)

    #
    # VM API
    #
    def get_vm(self, at_header: BlockHeaderAPI = None) -> VirtualMachineAPI:
        header = self.ensure_header(at_header)
        vm_class = self.get_vm_class_for_block_number(header.block_number)
        chain_context = ChainContext(self.chain_id)

        return vm_class(
            header=header,
            chaindb=self.chaindb,
            chain_context=chain_context,
            consensus_context=self.consensus_context,
        )

    #
    # Header API
    #
    def create_header_from_parent(
        self, parent_header: BlockHeaderAPI, **header_params: HeaderParams
    ) -> BlockHeaderAPI:
        return self.get_vm_class_for_block_number(
            block_number=BlockNumber(parent_header.block_number + 1),
        ).create_header_from_parent(parent_header, **header_params)

    def get_block_header_by_hash(self, block_hash: Hash32) -> BlockHeaderAPI:
        validate_word(block_hash, title="Block Hash")
        return self.chaindb.get_block_header_by_hash(block_hash)

    def get_canonical_block_header_by_number(
        self, block_number: BlockNumber
    ) -> BlockHeaderAPI:
        return self.chaindb.get_canonical_block_header_by_number(block_number)

    def get_canonical_head(self) -> BlockHeaderAPI:
        return self.chaindb.get_canonical_head()

    def get_score(self, block_hash: Hash32) -> int:
        return self.headerdb.get_score(block_hash)

    def ensure_header(self, header: BlockHeaderAPI = None) -> BlockHeaderAPI:
        """
        Return ``header`` if it is not ``None``, otherwise return the header
        of the canonical head.
        """
        if header is None:
            head = self.get_canonical_head()
            return self.create_header_from_parent(head)
        else:
            return header

    #
    # Block API
    #
    def get_ancestors(self, limit: int, header: BlockHeaderAPI) -> Tuple[BlockAPI, ...]:
        ancestor_count = min(header.block_number, limit)

        # We construct a temporary block object
        vm_class = self.get_vm_class_for_block_number(header.block_number)
        block_class = vm_class.get_block_class()
        block = block_class(header=header, uncles=[], transactions=[])

        ancestor_generator = iterate(
            compose(
                self.get_block_by_hash,
                operator.attrgetter("parent_hash"),
                operator.attrgetter("header"),
            ),
            block,
        )
        # we peel off the first element from the iterator which will be the
        # temporary block object we constructed.
        next(ancestor_generator)

        return tuple(take(ancestor_count, ancestor_generator))

    def get_block(self) -> BlockAPI:
        return self.get_vm().get_block()

    def get_block_by_hash(self, block_hash: Hash32) -> BlockAPI:
        validate_word(block_hash, title="Block Hash")
        block_header = self.get_block_header_by_hash(block_hash)
        return self.get_block_by_header(block_header)

    def get_block_by_header(self, block_header: BlockHeaderAPI) -> BlockAPI:
        vm = self.get_vm(block_header)
        return vm.get_block()

    def get_canonical_block_by_number(self, block_number: BlockNumber) -> BlockAPI:
        validate_uint256(block_number, title="Block Number")
        return self.get_block_by_hash(
            self.chaindb.get_canonical_block_hash(block_number)
        )

    def get_canonical_block_hash(self, block_number: BlockNumber) -> Hash32:
        return self.chaindb.get_canonical_block_hash(block_number)

    def build_block_with_transactions_and_withdrawals(
        self,
        transactions: Sequence[SignedTransactionAPI],
        parent_header: BlockHeaderAPI = None,
        withdrawals: Sequence[WithdrawalAPI] = None,
    ) -> Tuple[BlockAPI, Tuple[ReceiptAPI, ...], Tuple[ComputationAPI, ...]]:
        base_header = self.ensure_header(parent_header)
        vm = self.get_vm(base_header)

        new_header, receipts, computations = vm.apply_all_transactions(
            transactions,
            base_header,
        )

        if withdrawals:
            vm.apply_all_withdrawals(withdrawals)

        new_block = vm.set_block_transactions_and_withdrawals(
            vm.get_block(),
            new_header,
            transactions,
            receipts,
            withdrawals=withdrawals,
        )

        return new_block, receipts, computations

    #
    # Transaction API
    #
    def get_canonical_transaction_index(
        self, transaction_hash: Hash32
    ) -> Tuple[BlockNumber, int]:
        return self.chaindb.get_transaction_index(transaction_hash)

    def get_canonical_transaction(
        self, transaction_hash: Hash32
    ) -> SignedTransactionAPI:
        (block_num, index) = self.chaindb.get_transaction_index(transaction_hash)

        transaction = self.get_canonical_transaction_by_index(block_num, index)

        if transaction.hash == transaction_hash:
            return transaction
        else:
            raise TransactionNotFound(
                f"Found transaction {encode_hex(transaction.hash)} "
                f"instead of {encode_hex(transaction_hash)} in block "
                f"{block_num} at {index}"
            )

    def get_canonical_transaction_by_index(
        self, block_number: BlockNumber, index: int
    ) -> SignedTransactionAPI:
        VM_class = self.get_vm_class_for_block_number(block_number)

        return self.chaindb.get_transaction_by_index(
            block_number,
            index,
            VM_class.get_transaction_builder(),
        )

    def create_transaction(self, *args: Any, **kwargs: Any) -> SignedTransactionAPI:
        return self.get_vm().create_transaction(*args, **kwargs)

    def create_unsigned_transaction(
        self,
        *,
        nonce: int,
        gas_price: int,
        gas: int,
        to: Address,
        value: int,
        data: bytes,
    ) -> UnsignedTransactionAPI:
        return self.get_vm().create_unsigned_transaction(
            nonce=nonce,
            gas_price=gas_price,
            gas=gas,
            to=to,
            value=value,
            data=data,
        )

    def get_transaction_receipt(self, transaction_hash: Hash32) -> ReceiptAPI:
        (
            transaction_block_number,
            transaction_index,
        ) = self.chaindb.get_transaction_index(
            transaction_hash,
        )
        return self.get_transaction_receipt_by_index(
            transaction_block_number, transaction_index
        )

    def get_transaction_receipt_by_index(
        self, block_number: BlockNumber, index: int
    ) -> ReceiptAPI:
        vm = self.get_vm_class_for_block_number(block_number)

        receipt = self.chaindb.get_receipt_by_index(
            block_number,
            index,
            vm.get_receipt_builder(),
        )

        return receipt

    #
    # Execution API
    #
    def get_transaction_result(
        self, transaction: SignedTransactionAPI, at_header: BlockHeaderAPI
    ) -> bytes:
        with self.get_vm(at_header).in_costless_state() as state:
            computation = state.costless_execute_transaction(transaction)

        computation.raise_if_error()
        return computation.output

    def estimate_gas(
        self, transaction: SignedTransactionAPI, at_header: BlockHeaderAPI = None
    ) -> int:
        if at_header is None:
            at_header = self.get_canonical_head()
        with self.get_vm(at_header).in_costless_state() as state:
            return self.gas_estimator(state, transaction)

    def import_block(
        self, block: BlockAPI, perform_validation: bool = True
    ) -> BlockImportResult:
        try:
            parent_header = self.get_block_header_by_hash(block.header.parent_hash)
        except HeaderNotFound:
            raise ValidationError(
                f"Attempt to import block #{block.number}.  "
                f"Cannot import block {block.hash!r} before importing "
                f"its parent block at {block.header.parent_hash!r}"
            )

        base_header_for_import = self.create_header_from_parent(parent_header)
        # Make a copy of the empty header, adding in the expected amount of gas used.
        #   This allows for richer logging in the VM.
        annotated_header = base_header_for_import.copy(gas_used=block.header.gas_used)
        block_result = self.get_vm(annotated_header).import_block(block)
        imported_block = block_result.block

        # Validate the imported block.
        if perform_validation:
            try:
                validate_imported_block_unchanged(imported_block, block)
            except ValidationError:
                self.logger.warning(
                    "Proposed %s doesn't follow EVM rules, rejecting...", block
                )
                raise

        persist_result = self.persist_block(imported_block, perform_validation)
        return BlockImportResult(*persist_result, block_result.meta_witness)

    def persist_block(
        self, block: BlockAPI, perform_validation: bool = True
    ) -> BlockPersistResult:
        if perform_validation:
            self.validate_block(block)

        (
            new_canonical_hashes,
            old_canonical_hashes,
        ) = self.chaindb.persist_block(block)

        self.logger.debug(
            "Persisted block: number %s | hash %s",
            block.number,
            encode_hex(block.hash),
        )

        new_canonical_blocks = tuple(
            self.get_block_by_hash(header_hash) for header_hash in new_canonical_hashes
        )
        old_canonical_blocks = tuple(
            self.get_block_by_hash(header_hash) for header_hash in old_canonical_hashes
        )

        return BlockPersistResult(
            imported_block=block,
            new_canonical_blocks=new_canonical_blocks,
            old_canonical_blocks=old_canonical_blocks,
        )

    #
    # Validation API
    #
    def validate_receipt(self, receipt: ReceiptAPI, at_header: BlockHeaderAPI) -> None:
        VM_class = self.get_vm_class(at_header)
        VM_class.validate_receipt(receipt)

    def validate_block(self, block: BlockAPI) -> None:
        if block.is_genesis:
            raise ValidationError("Cannot validate genesis block this way")
        vm = self.get_vm(block.header)
        parent_header = self.get_block_header_by_hash(block.header.parent_hash)
        vm.validate_header(block.header, parent_header)
        vm.validate_seal(block.header)
        vm.validate_seal_extension(block.header, ())
        self.validate_uncles(block)

    def validate_seal(self, header: BlockHeaderAPI) -> None:
        vm = self.get_vm(header)
        vm.validate_seal(header)

    def validate_uncles(self, block: BlockAPI) -> None:
        has_uncles = len(block.uncles) > 0
        should_have_uncles = block.header.uncles_hash != EMPTY_UNCLE_HASH

        if not has_uncles and not should_have_uncles:
            # optimization to avoid loading ancestors from DB,
            # since the block has no uncles
            return
        elif has_uncles and not should_have_uncles:
            raise ValidationError(
                "Block has uncles but header suggests uncles should be empty"
            )
        elif should_have_uncles and not has_uncles:
            raise ValidationError(
                "Header suggests block should have uncles but block has none"
            )

        # Check for duplicates
        uncle_groups = groupby(operator.attrgetter("hash"), block.uncles)
        duplicate_uncles = tuple(
            sorted(hash for hash, twins in uncle_groups.items() if len(twins) > 1)
        )
        if duplicate_uncles:
            raise ValidationError(
                "Block contains duplicate uncles:\n"
                f" - {' - '.join(duplicate_uncles)}"
            )

        recent_ancestors = tuple(
            ancestor
            for ancestor in self.get_ancestors(MAX_UNCLE_DEPTH + 1, header=block.header)
        )
        recent_ancestor_hashes = {ancestor.hash for ancestor in recent_ancestors}
        recent_uncle_hashes = _extract_uncle_hashes(recent_ancestors)

        for uncle in block.uncles:
            if uncle.hash == block.hash:
                raise ValidationError("Uncle has same hash as block")

            # ensure the uncle has not already been included.
            if uncle.hash in recent_uncle_hashes:
                raise ValidationError(f"Duplicate uncle: {encode_hex(uncle.hash)}")

            # ensure that the uncle is not one of the canonical chain blocks.
            if uncle.hash in recent_ancestor_hashes:
                raise ValidationError(
                    f"Uncle {encode_hex(uncle.hash)} cannot be an ancestor "
                    f"of {encode_hex(block.hash)}"
                )

            # ensure that the uncle was built off of one of the canonical chain
            # blocks.
            if uncle.parent_hash not in recent_ancestor_hashes or (
                uncle.parent_hash == block.header.parent_hash
            ):
                raise ValidationError(
                    f"Uncle's parent {encode_hex(uncle.parent_hash)} "
                    f"is not an ancestor of {encode_hex(block.hash)}"
                )

            # Now perform VM level validation of the uncle
            self.validate_seal(uncle)

            try:
                uncle_parent = self.get_block_header_by_hash(uncle.parent_hash)
            except HeaderNotFound:
                raise ValidationError(
                    f"Uncle ancestor not found: {uncle.parent_hash!r}"
                )

            uncle_vm_class = self.get_vm_class_for_block_number(uncle.block_number)
            uncle_vm_class.validate_uncle(block, uncle, uncle_parent)


@to_set
def _extract_uncle_hashes(blocks: Iterable[BlockAPI]) -> Iterable[Hash32]:
    for block in blocks:
        for uncle in block.uncles:
            yield uncle.hash


class MiningChain(Chain, MiningChainAPI):
    header: BlockHeaderAPI = None

    def __init__(
        self, base_db: AtomicDatabaseAPI, header: BlockHeaderAPI = None
    ) -> None:
        super().__init__(base_db)
        self.header = self.ensure_header(header)

    def apply_transaction(
        self,
        transaction: SignedTransactionAPI,
    ) -> Tuple[BlockAPI, ReceiptAPI, ComputationAPI]:
        vm = self.get_vm(self.header)
        base_block = vm.get_block()

        receipt, computation = vm.apply_transaction(base_block.header, transaction)
        header_with_receipt = vm.add_receipt_to_header(base_block.header, receipt)

        # since we are building the block locally,
        # we have to persist all the incremental state
        vm.state.persist()
        new_header: BlockHeaderAPI = header_with_receipt.copy(
            state_root=vm.state.state_root
        )

        transactions = base_block.transactions + (transaction,)
        receipts = base_block.get_receipts(self.chaindb) + (receipt,)

        new_block = vm.set_block_transactions_and_withdrawals(
            base_block,
            new_header,
            transactions,
            receipts,
        )

        self.header = new_block.header

        return new_block, receipt, computation

    def import_block(
        self, block: BlockAPI, perform_validation: bool = True
    ) -> BlockImportResult:
        result = super().import_block(block, perform_validation)

        self.header = self.ensure_header()
        return result

    def set_header_timestamp(self, timestamp: int) -> None:
        self.header = self.header.copy(timestamp=timestamp)

    @staticmethod
    def _custom_header(base_header: BlockHeaderAPI, **kwargs: Any) -> BlockHeaderAPI:
        header_fields = {"coinbase"}
        header_params = {k: v for k, v in kwargs.items() if k in header_fields}
        return base_header.copy(**header_params)

    def mine_all(
        self,
        transactions: Sequence[SignedTransactionAPI],
        *args: Any,
        parent_header: BlockHeaderAPI = None,
        withdrawals: Sequence[WithdrawalAPI] = None,
        **kwargs: Any,
    ) -> Tuple[BlockImportResult, Tuple[ReceiptAPI, ...], Tuple[ComputationAPI, ...]]:
        if parent_header is None:
            base_header = self.header
        else:
            base_header = self.create_header_from_parent(parent_header)

        custom_header = self._custom_header(base_header, **kwargs)
        vm = self.get_vm(custom_header)

        new_header, receipts, computations = vm.apply_all_transactions(
            transactions, base_header
        )

        if withdrawals:
            vm.apply_all_withdrawals(withdrawals)

        filled_block = vm.set_block_transactions_and_withdrawals(
            vm.get_block(),
            new_header,
            transactions,
            receipts,
            withdrawals=withdrawals,
        )

        block_result = vm.mine_block(filled_block, *args, **kwargs)
        imported_block = block_result.block

        block_persist_result = self.persist_block(imported_block)
        block_import_result = BlockImportResult(
            *block_persist_result,
            block_result.meta_witness,
        )

        self.header = self.create_header_from_parent(imported_block.header)
        return (block_import_result, receipts, computations)

    def mine_block(self, *args: Any, **kwargs: Any) -> BlockAPI:
        """
        Mine whatever transactions have been incrementally applied so far.
        """
        return self.mine_block_extended(*args, **kwargs).block

    def mine_block_extended(self, *args: Any, **kwargs: Any) -> BlockAndMetaWitness:
        custom_header = self._custom_header(self.header, **kwargs)
        vm = self.get_vm(custom_header)
        current_block = vm.get_block()
        mine_result = vm.mine_block(current_block, *args, **kwargs)
        mined_block = mine_result.block

        self.validate_block(mined_block)

        self.chaindb.persist_block(mined_block)
        self.header = self.create_header_from_parent(mined_block.header)
        return mine_result

    def get_vm(self, at_header: BlockHeaderAPI = None) -> VirtualMachineAPI:
        if at_header is None:
            at_header = self.header

        return super().get_vm(at_header)



from typing import (
    Iterable,
)

from eth_typing import (
    Address,
    Hash32,
)
from eth_utils import (
    ValidationError,
    encode_hex,
    get_extended_debug_logger,
)
import lru

from eth.abc import (
    BlockHeaderAPI,
    ChainDatabaseAPI,
)
from eth.exceptions import (
    HeaderNotFound,
)

from ._utils import (
    get_block_signer,
    get_signers_at_checkpoint,
    is_checkpoint,
)
from .constants import (
    IN_MEMORY_SNAPSHOTS,
    NONCE_AUTH,
    NONCE_DROP,
)
from .datatypes import (
    MutableSnapshot,
    Snapshot,
    Tally,
    Vote,
    VoteAction,
)
from .encoding import (
    decode_snapshot,
    encode_snapshot,
)
from .exceptions import (
    SnapshotNotFound,
)


def make_snapshot_lookup_key(block_hash: Hash32) -> bytes:
    return b"block-hash-to-snapshot:%s" % block_hash


class SnapshotManager:
    """
    The ``SnapshotManager`` is responsible for managing the snapshots that hold the
    current state of the consensus engine. It creates new snapshots by applying headers
    on top of existing snapshots. It comes with APIs to create,
    persist and retrieve snapshots.
    """

    logger = get_extended_debug_logger(
        "eth.consensus.clique.snapshot_manager.SnapshotManager"
    )

    def __init__(self, chain_db: ChainDatabaseAPI, epoch_length: int) -> None:
        self._chain_db = chain_db
        self._epoch_length = epoch_length
        self._snapshots = lru.LRU(IN_MEMORY_SNAPSHOTS)

    def _lookup_header(
        self, block_hash: Hash32, parents: Iterable[BlockHeaderAPI]
    ) -> BlockHeaderAPI:
        for parent in parents:
            if parent.hash == block_hash:
                return parent
        try:
            return self._chain_db.get_block_header_by_hash(block_hash)
        except HeaderNotFound:
            raise ValidationError("Unknown ancestor %s", encode_hex(block_hash))

    def _create_snapshot_from_checkpoint_header(
        self, header: BlockHeaderAPI
    ) -> Snapshot:
        signers = get_signers_at_checkpoint(header)
        self.logger.debug2("Created snapshot from checkpoint at %s", header)

        snapshot = MutableSnapshot(
            signers=list(signers), block_hash=header.hash, votes=[], tallies={}
        )
        return self.add_snapshot(snapshot)

    def apply(self, current_snapshot: Snapshot, header: BlockHeaderAPI) -> Snapshot:
        """
        Apply the given header on top of the current snapshot to create a new snapshot.
        """

        if is_checkpoint(header.block_number, self._epoch_length):
            return self._create_snapshot_from_checkpoint_header(header)

        snapshot = current_snapshot.get_mutable_clone(header.hash)

        if header.nonce in {NONCE_AUTH, NONCE_DROP}:
            signer = get_block_signer(header)
            # Clear any votes from the signer regarding the subject that is voted on
            for vote in snapshot.votes:
                if vote.signer == signer and vote.subject == header.coinbase:
                    self.retract_vote(snapshot, vote.subject, vote.action)
                    snapshot.votes.remove(vote)
                    break

            action = VoteAction(header.nonce)

            if self.cast_vote(snapshot, header.coinbase, action):
                vote = Vote(
                    signer=signer,
                    block_number=header.block_number,
                    subject=header.coinbase,
                    action=action,
                )
                snapshot.votes.append(vote)

                tally = snapshot.tallies[header.coinbase]
                if tally.votes > len(snapshot.signers) / 2:
                    if tally.action is VoteAction.NOMINATE:
                        snapshot.signers.append(header.coinbase)
                        self.logger.debug("New signer added: %s", header.coinbase)
                    else:
                        if header.coinbase in snapshot.signers:
                            snapshot.signers.remove(header.coinbase)
                            self.logger.debug("Signer removed: %s", header.coinbase)

                    for vote in snapshot.votes.copy():
                        # Discard any pending votes *from* the added or removed member
                        if vote.signer == header.coinbase:
                            if self.retract_vote(snapshot, vote.subject, vote.action):
                                snapshot.votes.remove(vote)

                        # Discard any pending votes *regarding* the added or
                        # removed member. No need to uncast, the whole tally is going
                        # to be removed anyway.
                        if vote.subject == header.coinbase:
                            snapshot.votes.remove(vote)

                    snapshot.tallies.pop(header.coinbase)

        self.add_snapshot(snapshot)

        return snapshot.get_immutable_clone()

    def get_or_create_snapshot(
        self,
        block_number: int,
        block_hash: Hash32,
        parents: Iterable[BlockHeaderAPI] = (),
    ) -> Snapshot:
        """
        Return a snapshot either by creating or retrieving it or raise a
        ``ValidationError`` if the header does not have a known ancestor.
        """
        try:
            snapshot = self.get_snapshot(block_number, block_hash)
        except SnapshotNotFound:
            return self.create_snapshot_for(block_hash, parents)
        else:
            return snapshot

    def create_snapshot_for(
        self, block_hash: Hash32, cached_parents: Iterable[BlockHeaderAPI]
    ) -> Snapshot:
        """
        Create a ``Snapshot`` for the given ``block_hash``. This involves traversing
        backwards through the chain of headers to find a suitable base snapshot either
        from memory, on disk or by creating it on the fly from a checkpoint header.
        After we've found a suitable base snapshot, apply all headers from after the
        base snapshot up to the header of ``block_hash`` to create the requested
        snapshot.
        """
        current_header = header = self._lookup_header(block_hash, cached_parents)

        if is_checkpoint(header.block_number, self._epoch_length):
            return self._create_snapshot_from_checkpoint_header(current_header)

        parents = []
        while True:
            try:
                new_snapshot = self.get_snapshot(
                    current_header.block_number, current_header.parent_hash
                )
            except SnapshotNotFound:
                current_header = self._lookup_header(
                    current_header.parent_hash, cached_parents
                )

                if is_checkpoint(current_header.block_number, self._epoch_length):
                    new_snapshot = self._create_snapshot_from_checkpoint_header(
                        current_header
                    )
                    break
                else:
                    parents.append(current_header)
            else:
                break

        for parent in reversed(parents):
            new_snapshot = self.apply(new_snapshot, parent)

        new_snapshot = self.apply(new_snapshot, header)

        if is_checkpoint(header.block_number, self._epoch_length):
            self.logger.debug2(
                "Persisting checkpoint snapshot at %s", header.block_number
            )
            self.persist_snapshot(new_snapshot)

        return new_snapshot

    def get_snapshot(self, block_number: int, block_hash: Hash32) -> Snapshot:
        """
        Return a ``Snapshot`` if it exists in memory, on-disk or can be computed
        directly from a header that serves as a checkpoint.
        Otherwise raise a ``SnapshotNotFound`` error.
        """

        # We first try to find the snapshot in memory
        if block_hash in self._snapshots:
            return self._snapshots[block_hash]

        if is_checkpoint(block_number, self._epoch_length):
            try:
                # We might have it saved on disk
                return self.get_snapshot_from_db(block_hash)
            except SnapshotNotFound:
                pass

            try:
                # Otherwise, we can retrieve it on the fly
                header = self._chain_db.get_block_header_by_hash(block_hash)
            except HeaderNotFound:
                raise SnapshotNotFound(
                    f"Can not get snapshot for {block_hash!r} at {block_number}"
                )
            else:
                if header.block_number != block_number:
                    raise SnapshotNotFound(
                        f"Can not get snapshot for {block_hash!r} at {block_number}"
                    )
                else:
                    return self._create_snapshot_from_checkpoint_header(header)

        raise SnapshotNotFound(
            f"Can not get snapshot for {block_hash!r} at {block_number}"
        )

    def add_snapshot(self, mutable_snapshot: MutableSnapshot) -> Snapshot:
        """
        Retrieve a ``Snapshot`` from the given ``mutable_snapshot``
        and add it to the local cache.
        """
        snapshot = mutable_snapshot.get_immutable_clone()
        self._snapshots[snapshot.block_hash] = snapshot
        return snapshot

    def persist_snapshot(self, snapshot: Snapshot) -> None:
        """
        Persist the given snapshot to the database.
        """
        key = make_snapshot_lookup_key(snapshot.block_hash)
        with self._chain_db.db.atomic_batch() as db:
            db[key] = encode_snapshot(snapshot)

    def get_snapshot_from_db(self, block_hash: Hash32) -> Snapshot:
        """
        Retrieve a snapshot from the database.
        Raise ``SnapshotNotFound`` if it does not exist.
        """
        key = make_snapshot_lookup_key(block_hash)
        try:
            encoded_key = self._chain_db.db[key]
        except KeyError as e:
            raise SnapshotNotFound(
                f"Can not get on-disk snapshot for {block_hash!r}"
            ) from e
        else:
            return decode_snapshot(encoded_key)

    def cast_vote(
        self, snapshot: MutableSnapshot, subject: Address, action: VoteAction
    ) -> bool:
        """
        Cast a vote on a ``MutableSnapshot``.
        """
        try:
            action.validate_for(snapshot.signers, subject)
        except ValidationError:
            return False

        if subject not in snapshot.tallies:
            snapshot.tallies[subject] = Tally(action)

        snapshot.tallies[subject] = snapshot.tallies[subject].upvote()
        return True

    def retract_vote(
        self, snapshot: MutableSnapshot, subject: Address, action: VoteAction
    ) -> bool:
        """
        Retract a vote on a ``MutableSnapshot``.
        """
        if subject not in snapshot.tallies:
            # Dangling votes are simply dropped
            return True

        tally = snapshot.tallies[subject]

        if tally.action is not action:
            return False

        if tally.votes > 1:
            snapshot.tallies[subject] = tally.downvote()
        else:
            snapshot.tallies.pop(subject)

        return True



from abc import (
    ABC,
)

from eth_keys.datatypes import (
    PrivateKey,
)

from eth._utils.transactions import (
    create_transaction_signature,
)
from eth.vm.forks.paris.transactions import (
    ParisLegacyTransaction,
    ParisTransactionBuilder,
    ParisUnsignedLegacyTransaction,
)


class ShanghaiLegacyTransaction(ParisLegacyTransaction, ABC):
    pass


class ShanghaiUnsignedLegacyTransaction(ParisUnsignedLegacyTransaction):
    def as_signed_transaction(
        self, private_key: PrivateKey, chain_id: int = None
    ) -> ShanghaiLegacyTransaction:
        v, r, s = create_transaction_signature(self, private_key, chain_id=chain_id)
        return ShanghaiLegacyTransaction(
            nonce=self.nonce,
            gas_price=self.gas_price,
            gas=self.gas,
            to=self.to,
            value=self.value,
            data=self.data,
            v=v,
            r=r,
            s=s,
        )


class ShanghaiTransactionBuilder(ParisTransactionBuilder):
    legacy_signed = ShanghaiLegacyTransaction
    legacy_unsigned = ShanghaiUnsignedLegacyTransaction



from contextlib import (
    contextmanager,
)
from typing import (
    Iterator,
)
import warnings


# TODO: drop once https://github.com/cython/cython/issues/1720 is resolved
@contextmanager
def catch_and_ignore_import_warning() -> Iterator[None]:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ImportWarning)
        yield



from eth_utils.exceptions import (
    ValidationError,
)

from eth.abc import (
    SignedTransactionAPI,
    StateAPI,
)
from eth.vm.forks.homestead.validation import (
    validate_homestead_transaction,
)


def validate_london_normalized_transaction(
    state: StateAPI,
    transaction: SignedTransactionAPI,
) -> None:
    """
    Validates a London normalized transaction.

    Raise `eth.exceptions.ValidationError` if the sender cannot
    afford to send this transaction.
    """
    base_fee_per_gas = state.execution_context.base_fee_per_gas
    if transaction.max_fee_per_gas < base_fee_per_gas:
        raise ValidationError(
            f"Sender's max fee per gas ({transaction.max_fee_per_gas}) is "
            f"lower than block's base fee per gas ({base_fee_per_gas})"
        )

    validate_homestead_transaction(state, transaction)



from typing import (
    Type,
)

from eth_bloom import (
    BloomFilter,
)

from eth.abc import (
    BlockAPI,
    BlockHeaderAPI,
    ReceiptAPI,
    StateAPI,
    SignedTransactionAPI,
    ComputationAPI,
)
from eth.constants import (
    BLOCK_REWARD,
    UNCLE_DEPTH_PENALTY_FACTOR,
    ZERO_HASH32,
)
from eth.rlp.logs import Log
from eth.rlp.receipts import Receipt

from eth.vm.base import VM

from .blocks import FrontierBlock
from .constants import MAX_REFUND_QUOTIENT
from .state import FrontierState
from .headers import (
    create_frontier_header_from_parent,
    compute_frontier_difficulty,
    configure_frontier_header,
)
from .validation import validate_frontier_transaction_against_header


def make_frontier_receipt(
    computation: ComputationAPI, new_cumulative_gas_used: int
) -> ReceiptAPI:
    # Reusable for other forks
    # This skips setting the state root (set to 0 instead). The logic for making a
    # state root lives in the FrontierVM, so that state merkelization at each receipt
    # is skipped at Byzantium+.

    logs = [
        Log(address, topics, data)
        for address, topics, data in computation.get_log_entries()
    ]

    receipt = Receipt(
        state_root=ZERO_HASH32,
        gas_used=new_cumulative_gas_used,
        logs=logs,
    )

    return receipt


class FrontierVM(VM):
    # fork name
    fork: str = "frontier"

    # classes
    block_class: Type[BlockAPI] = FrontierBlock
    _state_class: Type[StateAPI] = FrontierState

    # methods
    create_header_from_parent = staticmethod(create_frontier_header_from_parent)  # type: ignore  # noqa: E501
    compute_difficulty = staticmethod(compute_frontier_difficulty)  # type: ignore
    configure_header = configure_frontier_header
    validate_transaction_against_header = validate_frontier_transaction_against_header

    @staticmethod
    def get_block_reward() -> int:
        return BLOCK_REWARD

    @staticmethod
    def get_uncle_reward(block_number: int, uncle: BlockHeaderAPI) -> int:
        return (
            BLOCK_REWARD
            * (UNCLE_DEPTH_PENALTY_FACTOR + uncle.block_number - block_number)
            // UNCLE_DEPTH_PENALTY_FACTOR
        )

    @classmethod
    def get_nephew_reward(cls) -> int:
        return cls.get_block_reward() // 32

    def add_receipt_to_header(
        self, old_header: BlockHeaderAPI, receipt: ReceiptAPI
    ) -> BlockHeaderAPI:
        return old_header.copy(
            bloom=int(BloomFilter(old_header.bloom) | receipt.bloom),
            gas_used=receipt.gas_used,
            state_root=self.state.make_state_root(),
        )

    @classmethod
    def calculate_net_gas_refund(cls, consumed_gas: int, gross_refund: int) -> int:
        max_refund = consumed_gas // MAX_REFUND_QUOTIENT
        return min(max_refund, gross_refund)

    @classmethod
    def finalize_gas_used(
        cls, transaction: SignedTransactionAPI, computation: ComputationAPI
    ) -> int:
        gas_remaining = computation.get_gas_remaining()
        consumed_gas = transaction.gas - gas_remaining

        gross_refund = computation.get_gas_refund()
        net_refund = cls.calculate_net_gas_refund(consumed_gas, gross_refund)

        return consumed_gas - net_refund

    @classmethod
    def make_receipt(
        cls,
        base_header: BlockHeaderAPI,
        transaction: SignedTransactionAPI,
        computation: ComputationAPI,
        state: StateAPI,
    ) -> ReceiptAPI:
        gas_used = base_header.gas_used + cls.finalize_gas_used(
            transaction, computation
        )

        receipt_without_state_root = make_frontier_receipt(computation, gas_used)

        return receipt_without_state_root.copy(state_root=state.make_state_root())



from collections.abc import (
    Mapping,
    MutableMapping,
)
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterable,
    Tuple,
    Union,
    cast,
)

from eth_utils import (
    encode_hex,
    to_tuple,
)

from eth.abc import (
    DatabaseAPI,
)
from eth.vm.interrupt import (
    EVMMissingData,
)

if TYPE_CHECKING:
    ABC_Mutable_Mapping = MutableMapping[bytes, Union[bytes, "MissingReason"]]
    ABC_Mapping = Mapping[bytes, Union[bytes, "MissingReason"]]
else:
    ABC_Mutable_Mapping = MutableMapping
    ABC_Mapping = Mapping


class MissingReason(str):
    pass


NEVER_INSERTED = MissingReason("Key is missing because it was never inserted")
DELETED = MissingReason("Key is missing because it was deleted")


class DiffMissingError(KeyError):
    """
    Raised when trying to access a missing key/value pair in a :class:`DBDiff`
    or :class:`DBDiffTracker`.

    Use :attr:`is_deleted` to check if the value is missing because it was
    deleted, or simply because it was never updated.
    """

    def __init__(self, missing_key: bytes, reason: MissingReason) -> None:
        self.reason = reason
        super().__init__(missing_key, reason)

    @property
    def is_deleted(self) -> bool:
        return self.reason == DELETED


class DBDiffTracker(ABC_Mutable_Mapping):
    """
    Records changes to a :class:`~eth.abc.DatabaseAPI`

    If no value is available for a key, it could be for one of two reasons:
    - the key was never updated during tracking
    - the key was deleted at some point

    When getting a value, a special subtype of KeyError is raised on failure.
    The exception, :class:`DiffMissingError`, can be used to check if the value
    was deleted, or never present, using :meth:`DiffMissingError.is_deleted`.

    When it's time to take the tracked changes and write them to your database,
    get the :class:`DBDiff` with :meth:`DBDiffTracker.diff` and use the attached
    methods.
    """

    def __init__(self) -> None:
        self._changes: Dict[bytes, Union[bytes, MissingReason]] = {}

    def __contains__(self, key: bytes) -> bool:  # type: ignore # Breaks LSP
        result = self._changes.get(key, NEVER_INSERTED)
        return result not in (DELETED, NEVER_INSERTED)

    def __getitem__(self, key: bytes) -> bytes:
        result = self._changes.get(key, NEVER_INSERTED)
        if result in (DELETED, NEVER_INSERTED):
            raise DiffMissingError(key, result)  # type: ignore # ignore over cast for perf reasons  # noqa: E501
        else:
            return result  # type: ignore # ignore over cast for perf reasons

    def __setitem__(self, key: bytes, value: Union[bytes, MissingReason]) -> None:
        self._changes[key] = value

    def __delitem__(self, key: bytes) -> None:
        # The diff does not have access to any underlying db,
        # so it cannot check if the key exists before deleting.
        self._changes[key] = DELETED

    def __iter__(self) -> None:
        raise NotImplementedError(
            "Cannot iterate through changes, use diff().apply_to(db) "
            "to update a database"
        )

    def __len__(self) -> int:
        return len(self._changes)

    def diff(self) -> "DBDiff":
        return DBDiff(dict(self._changes))


class DBDiff(ABC_Mapping):
    """
    DBDiff is a read-only view of the updates/inserts and deletes
    generated when tracking changes with :class:`DBDiffTracker`.

    The primary usage is to apply these changes to your underlying
    database with :meth:`apply_to`.
    """

    _changes: Dict[bytes, Union[bytes, MissingReason]] = None

    def __init__(
        self, changes: Dict[bytes, Union[bytes, MissingReason]] = None
    ) -> None:
        if changes is None:
            self._changes = {}
        else:
            self._changes = changes

    def __getitem__(self, key: bytes) -> bytes:
        result = self._changes.get(key, NEVER_INSERTED)
        if result in (DELETED, NEVER_INSERTED):
            raise DiffMissingError(key, result)  # type: ignore # ignore over cast for perf reasons  # noqa: E501
        else:
            return result  # type: ignore # ignore over cast for perf reasons

    def __iter__(self) -> None:
        raise NotImplementedError(
            "Cannot iterate through changes, use apply_to(db) to update a database. "
            "Also, pending_keys(), deleted_keys(), and pending_items() might be of "
            "interest."
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DBDiff):
            return False
        else:
            return self._changes == other._changes

    def __repr__(self) -> str:
        deleted = [
            f"key={encode_hex(key)}"
            for key, val in self._changes.items()
            if val is DELETED
        ]
        updated = [
            f"key={encode_hex(key)} to val={encode_hex(cast(bytes, val))}"
            for key, val in self._changes.items()
            if val is not DELETED
        ]
        return f"<DBDiff: deletions={deleted!r}, updates={updated!r}>"

    def __len__(self) -> int:
        return len(self._changes)

    @to_tuple
    def deleted_keys(self) -> Iterable[bytes]:
        """
        List all the keys that have been deleted.
        """
        for key, value in self._changes.items():
            if value is DELETED:
                yield key

    @to_tuple
    def pending_keys(self) -> Iterable[bytes]:
        """
        List all the keys who have had values change. This IGNORES
        any keys that have been deleted.
        """
        for key, value in self._changes.items():
            if value is not DELETED:
                yield key

    @to_tuple
    def pending_items(self) -> Iterable[Tuple[bytes, bytes]]:
        """
        A tuple of (key, value) pairs for every key that has been updated.
        Like :meth:`pending_keys()`, this does not return any deleted keys.
        """
        for key, value in self._changes.items():
            if value is not DELETED:
                yield key, value  # type: ignore # value can only be DELETED or actual new value  # noqa: E501

    def apply_to(
        self, db: Union[DatabaseAPI, ABC_Mutable_Mapping], apply_deletes: bool = True
    ) -> None:
        """
        Apply the changes in this diff to the given database.
        You may choose to opt out of deleting any underlying keys.

        :param apply_deletes: whether the pending deletes should be
            applied to the database
        """
        for key, value in self._changes.items():
            if value is DELETED:
                if apply_deletes:
                    try:
                        del db[key]
                    except EVMMissingData:
                        raise
                    except KeyError:
                        pass
                else:
                    pass
            else:
                db[key] = value  # type: ignore # ignore over cast for perf reasons

    @classmethod
    def join(cls, diffs: Iterable["DBDiff"]) -> "DBDiff":
        """
        Join several DBDiff objects into a single DBDiff object.

        In case of a conflict, changes in diffs that come later
        in ``diffs`` will overwrite changes from earlier changes.
        """
        tracker = DBDiffTracker()
        for diff in diffs:
            diff.apply_to(tracker)
        return tracker.diff()
